#!/usr/bin/env python

# container-service-extension
# Copyright (c) 2019 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

import os
import sys
import time

import click
import cryptography
import pyvcloud.vcd.client as vcd_client
from pyvcloud.vcd.exceptions import BadRequestException
from pyvcloud.vcd.exceptions import EntityNotFoundException
from pyvcloud.vcd.exceptions import VcdException
import requests
from vcd_cli.utils import stdout
import yaml

from container_service_extension.common.constants.server_constants import CONFIG_DECRYPTION_ERROR_MSG, LegacyLocalTemplateKey  # noqa: E501
from container_service_extension.common.constants.server_constants import LocalTemplateKey  # noqa: E501
from container_service_extension.common.constants.shared_constants import SYSTEM_ORG_NAME  # noqa: E501
import container_service_extension.common.utils.core_utils as utils
import container_service_extension.common.utils.pyvcloud_utils as vcd_utils
import container_service_extension.common.utils.server_utils as server_utils
from container_service_extension.installer.config_validator import get_validated_config  # noqa: E501
import container_service_extension.installer.configure_cse as configure_cse
from container_service_extension.installer.cse_service_role_mgr import create_cse_service_role  # noqa : E501
from container_service_extension.installer.sample_generator import generate_sample_config  # noqa: E501
import container_service_extension.installer.templates.local_template_manager as ltm  # noqa: E501
from container_service_extension.installer.templates.remote_template_manager import RemoteTemplateManager  # noqa: E501
from container_service_extension.lib.telemetry.constants import CseOperation
from container_service_extension.lib.telemetry.constants import OperationStatus
from container_service_extension.lib.telemetry.constants import PayloadKey
from container_service_extension.lib.telemetry.telemetry_handler \
    import record_user_action
from container_service_extension.lib.telemetry.telemetry_handler import \
    record_user_action_details
from container_service_extension.lib.telemetry.telemetry_utils \
    import store_telemetry_settings
from container_service_extension.logging.logger import INSTALL_LOGGER
from container_service_extension.logging.logger import INSTALL_WIRELOG_FILEPATH
from container_service_extension.logging.logger import NULL_LOGGER
from container_service_extension.logging.logger import SERVER_CLI_LOGGER
from container_service_extension.logging.logger import SERVER_CLI_WIRELOG_FILEPATH  # noqa: E501
from container_service_extension.logging.logger import SERVER_CLOUDAPI_WIRE_LOGGER  # noqa: E501
from container_service_extension.logging.logger import SERVER_DEBUG_WIRELOG_FILEPATH  # noqa: E501
from container_service_extension.logging.logger import SERVER_LOGGER
from container_service_extension.security.encryption_engine import decrypt_file
from container_service_extension.security.encryption_engine import encrypt_file
from container_service_extension.security.encryption_engine import get_decrypted_file_contents  # noqa: E501
import container_service_extension.server.service as cse_service

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

# Template display options
DISPLAY_ALL = "all"
DISPLAY_DIFF = "diff"
DISPLAY_LOCAL = "local"
DISPLAY_REMOTE = "remote"

# Prompt messages
PASSWORD_FOR_CONFIG_ENCRYPTION_MSG = "Password for config file encryption"
PASSWORD_FOR_CONFIG_DECRYPTION_MSG = "Password for config file decryption"
USERNAME_FOR_SYSTEM_ADMINISTRATOR = "Username for System Administrator"
PASSWORD_FOR_SYSTEM_ADMINISTRATOR = "Password for "


@click.group(context_settings=CONTEXT_SETTINGS, invoke_without_command=True)
@click.pass_context
def cli(ctx):
    """Container Service Extension for VMware vCloud Director.

\b
    Examples, by default, following commands expect an encrypted CSE
    configuration file. To use a plain-text configuration file instead, specify
    the flag --skip-config-decryption.
\b
        cse version
            Display the software version of the CSE server.
\b
        cse sample
            Generate sample CSE configuration and print it to the console.
\b
        cse sample --pks-config
            Generate sample PKS configuration and print it to the console.
\b
        cse sample --output config.yaml
            Generate sample CSE configuration, and write it to the file
            'config.yaml'.
\b
        cse sample --pks-config --output pks-config.yaml
            Generate sample PKS configuration, and write it to the file
            'pks-config.yaml'
\b
        cse install --config config.yaml --skip-config-decryption
            Install CSE using configuration specified in 'config.yaml'.
\b
        cse install -c encrypted-config.yaml
            Install CSE using configuration specified in 'encrypted-config.yaml'.
            The configuration file will be decrypted in memory using a password
            (user will be prompted for the password).
\b
        cse install -c encrypted-config.yaml --pks-config encrypted-pks-config.yaml
            Install CSE using configuration specified in 'encrypted-config.yaml'
            and 'encrypted-pks-config.yaml'. Both these files will be decrypted
            in memory using the same password (user will be prompted for the
            password).
\b
        cse install -c config.yaml --force-update --skip-config-decryption
            Install CSE, and if the templates already exist in vCD, create
            them again.
\b
        cse install -c config.yaml --retain-temp-vapp --ssh-key ~/.ssh/id_rsa.pub
            Install CSE, retain the temporary vApp after the templates have
            been captured. Copy specified SSH key into all template VMs so
            users with the corresponding private key have access (--ssh-key is
            required when --retain-temp-vapp is used).
\b
        cse upgrade --config config.yaml --skip-config-decryption
            Upgrade CSE using configuration specified in 'config.yaml'.
\b
        cse upgrade -c encrypted-config.yaml
            Upgrade CSE using configuration specified in
            'encrypted-config.yaml'. The configuration file will be decrypted
            in memory using a password (user will be prompted for the password).
\b
        cse upgrade -c config.yaml --retain-temp-vapp --ssh-key ~/.ssh/id_rsa.pub
            Upgrade CSE, retain the temporary vApp after the templates have
            been captured. Copy specified SSH key into all template VMs so
            users with the corresponding private key have access (--ssh-key is
            required when --retain-temp-vapp is used).
\b
        cse pks-configure --skip-config-decryption pks_config.yaml
            Configure CSE for PKS using pks_config.yaml. Skip decryption of
            the config information.
\b
        cse check config.yaml --skip-config-decryption
            Checks validity of 'config.yaml'.
\b
        cse check config.yaml --pks-config pks-config.yaml --skip-config-decryption
            Checks validity of both config.yaml and 'pks-config.yaml'.
\b
        cse check config.yaml --check-install --skip-config-decryption
            Checks validity of 'config.yaml'.
            Checks that CSE is installed on vCD according to 'config.yaml'
\b
        cse run -c encrypted-config.yaml
            Run CSE Server using configuration supplied via 'encrypted-config.yaml'.
            Will first validate that CSE was installed according to
            'encrypted-config.yaml'. User will be prompted for the password
            required for decrypting 'encrypted-config.yaml'.
\b
        cse run -c encrypted-config.yaml --pks-config encrypted-pks-config.yaml
            Run CSE Server using configuration specified in 'encrypted-config.yaml'
            and 'encrypted-pks-config.yaml'. Also validate that CSE has been
            installed according to 'encrypted-config.yaml' and PKS was installed
            according to 'encrypted-pks-config.yaml'. User will be prompted for
            the password required to decrypt both the configuration files.
\b
        cse run -c encrypted-config.yaml --skip-check
            Run CSE Server using configuration specified in 'encrypted-config.yaml'
            without first validating that CSE was installed according
            to 'encrypted-config.yaml'. User will be prompted for the password
            required to decrypt the configuration file.
\b
        cse run -c config.yaml --skip-check --skip-config-decryption
            Run CSE Server using configuration specified in plain text 'config.yaml'
            without first validating that CSE was installed according to 'config.yaml'.
\b
        cse encrypt config.yaml --output ~/.configs/encrypted-config.yaml
            Encrypt the plain text configuration file viz. config.yaml and save
            it the encrypted version at the specified location. If --output
            flag is not specified, the encrypted content will be displayed on
            console. User will be prompted to provide a password for encryption.
\b
        cse decrypt config.yaml --output ~./configs/encrypted-config.yaml
            Decrypt the configuration file encrypted-config.yaml and save the
            decrypted content in the file config.yaml. If --output flag is
            not specified the decrypted content will be printed onto console.
            User will be prompted for the password required for decryption.
\b
        cse template install mytemplate 1 -c myconfig.yaml
            Installs mytemplate at revision 1 specified in the remote template
            repository URL specified in myconfig.yaml
\b
        cse template install * * -c myconfig.yaml
            Installs all templates specified in the remote template
            repository URL specified in myconfig.yaml
\b
    Environment Variables
        CSE_CONFIG
            If this environment variable is set, the CSE commands will use the
            file specified in the environment variable as CSE configuration
            file. However if user provides a configuration file along with the
            command, it will override the environment variable. If neither the
            environment variable is set nor user provides a configuration file,
            CSE will try to read configuration from the file 'config.yaml' in
            the current directory.
\b
        CSE_CONFIG_PASSWORD
            If this environment variable is set, the commands will use the
            value provided in this variable to encrypt/decrypt CSE
            configuration file. If this variable is not set, then user will be
            prompted for password to encrypt/decrypt the CSE configuration files.
    """  # noqa: E501
    if ctx.invoked_subcommand is None:
        console_message_printer = utils.ConsoleMessagePrinter()
        console_message_printer.general_no_color(ctx.get_help())
        return


@cli.group(short_help='Manage native Kubernetes runtime templates')
@click.pass_context
def template(ctx):
    """Manage native Kubernetes runtime templates.

\b
    Examples, by default, following commands expect an encrypted CSE
    configuration file. To use a plain-text configuration file instead, specify
    the flag --skip-config-decryption.
\b
        cse template list -c encrypted-config.yaml
            Display all templates, including the ones that are currently in the local
            catalog, and the ones that are defined in remote template repository.
            'encrypted-config.yaml' will be decrypted with the password received
            from user via a console prompt.
\b
        cse template list -c config.yaml --skip-config-decryption
            Display all templates, including the ones that are currently in the local
            catalog, and the ones that are defined in remote template repository.
\b
        cse template list -c config.yaml --display local --skip-config-decryption
            Display templates that are currently in the local catalog.
\b
        cse template list -c config.yaml --display remote --skip-config-decryption
            Display templates that are defined in the remote template repository.
\b
        cse template list -c config.yaml --display diff --skip-config-decryption
            Display only templates that are defined in remote template repository
            but not present in the local catalog.
\b
        cse template install -c config.yaml --skip-config-decryption
            Install all templates defined in remote template repository that are
            missing from the local catalog.
\b
        cse template install -c encrypted-config.yaml
            Install all templates defined in remote template repository that are
            missing from the local catalog. 'encrypted-config.yaml' will be
            decrypted with the password received from user via a console prompt.
\b
        cse template install [template name] [template revision] -c encrypted-config.yaml
            Install a particular template at a given revision defined in remote
            template repository that is missing from the local catalog.
            'encrypted-config.yaml' will be decrypted with the password received
            from user via a console prompt.
\b
        cse template install -c config.yaml --force --skip-config-decryption
            Install all templates defined in remote template repository. Templates
            already present in the local catalog that match with one in the remote
            repository will be recreated from scratch.
    """  # noqa: E501
    pass


@cli.command(short_help='Display CSE version')
@click.pass_context
def version(ctx):
    """Display CSE version."""
    SERVER_CLI_LOGGER.debug(f"Executing command: {ctx.command_path}")
    cse_info = utils.get_cse_info()
    ver_str = '%s, %s, version %s' % (cse_info['product'],
                                      cse_info['description'],
                                      cse_info['version'])
    stdout(cse_info, ctx, ver_str)


@cli.command(short_help='Create CSE service role for CSE install/upgrade/run')
@click.pass_context
@click.argument('vcd_host', metavar='VCD_HOST')
@click.option(
    '-s',
    '--skip-verify-ssl-certs',
    is_flag=True,
    help='Skip verifying SSL certificates of VCD Host')
def create_service_role(ctx, vcd_host, skip_verify_ssl_certs):
    """Create CSE service Role."""
    SERVER_CLI_LOGGER.debug(f"Executing command: {ctx.command_path}")
    console_message_printer = utils.ConsoleMessagePrinter()
    requests.packages.urllib3.disable_warnings()

    # The console_message_printer is not being passed to the python version
    # check, because we want to suppress the version check messages from being
    # printed onto console.
    utils.check_python_version()

    # Prompt user for administrator username/password
    admin_username = utils.prompt_text(USERNAME_FOR_SYSTEM_ADMINISTRATOR,
                                       color='green', hide_input=False)
    admin_password = utils.prompt_text(PASSWORD_FOR_SYSTEM_ADMINISTRATOR + str(admin_username),  # noqa: E501
                                       color='green', hide_input=True)

    msg = f"Connecting to vCD: {vcd_host}"
    console_message_printer.general_no_color(msg)
    SERVER_CLI_LOGGER.info(msg)

    client = None
    try:
        client = vcd_client.Client(vcd_host, verify_ssl_certs=not skip_verify_ssl_certs)  # noqa: E501
        credentials = vcd_client.BasicLoginCredentials(
            admin_username,
            SYSTEM_ORG_NAME,
            admin_password)
        client.set_credentials(credentials)

        msg = f"Connected to vCD as system administrator: {admin_username}"
        console_message_printer.general_no_color(msg)
        SERVER_CLI_LOGGER.info(msg)
        msg = "Creating CSE Service Role..."
        console_message_printer.general_no_color(msg)
        SERVER_CLI_LOGGER.info(msg)

        try:
            create_cse_service_role(
                client,
                msg_update_callback=console_message_printer,
                logger_debug=SERVER_CLI_LOGGER)
        except EntityNotFoundException as err:
            msg = "CSE Internal Error, Please contact support"
            console_message_printer.error(msg)
            SERVER_CLI_LOGGER.error(msg)
            console_message_printer.error(str(err))
            SERVER_CLI_LOGGER.error(str(err))
        except BadRequestException as err:
            msg = "CSE Service Role already Exists"
            console_message_printer.error(msg)
            console_message_printer.error(str(err))
            SERVER_CLI_LOGGER.error(msg)
            SERVER_CLI_LOGGER.error(str(err))
    except requests.exceptions.ConnectionError as err:
        console_message_printer.error(str(err))
        SERVER_CLI_LOGGER.error(str(err))
    except VcdException as err:
        msg = f"Incorrect SystemOrg Username: {admin_username} and/or Password"
        console_message_printer.error(msg)
        console_message_printer.error(str(err))
        SERVER_CLI_LOGGER.error(msg)
        SERVER_CLI_LOGGER.error(str(err))
    except Exception as err:
        console_message_printer.error(str(err))
        SERVER_CLI_LOGGER.error(str(err))
    finally:
        if client is not None:
            client.logout()


@cli.command(short_help='Generate sample CSE/PKS configuration')
@click.pass_context
@click.option(
    '-o',
    '--output',
    'output_file_name',
    required=False,
    default=None,
    metavar='OUTPUT_FILE_NAME',
    help="Filepath to write configuration file to")
@click.option(
    '-p',
    '--pks-config',
    is_flag=True,
    help='Generate only sample PKS config')
@click.option(
    '-l',
    '--legacy-mode',
    'legacy_mode',
    required=False,
    is_flag=True,
    help="Generate sample config for legacy type")  # noqa: E501
def sample(ctx, output_file_name, pks_config, legacy_mode):
    """Display sample CSE config file contents."""
    SERVER_CLI_LOGGER.debug(f"Executing command: {ctx.command_path}")
    console_message_printer = utils.ConsoleMessagePrinter()
    # The console_message_printer is not being passed to the python version
    # check, because we want to suppress the version check messages from being
    # printed onto console, and pollute the sample config.
    utils.check_python_version()

    try:
        sample_config = generate_sample_config(
            output_file_name=output_file_name,
            generate_pks_config=pks_config,
            legacy_mode=legacy_mode)
    except Exception as err:
        console_message_printer.error(str(err))
        SERVER_CLI_LOGGER.error(str(err))
        sys.exit(1)

    console_message_printer.general_no_color(sample_config)
    SERVER_CLI_LOGGER.debug(sample_config)


@cli.command(short_help="Checks that CSE/PKS config file is valid. Can "
                        "also check that CSE/PKS is installed according to "
                        "the config file(s)")
@click.pass_context
@click.argument('config_file_path', metavar='CONFIG_FILE_NAME',
                envvar='CSE_CONFIG',
                default='config.yaml',
                type=click.Path(exists=True))
@click.option(
    '-p',
    '--pks-config-file',
    'pks_config_file_path',
    type=click.Path(exists=True),
    metavar='PKS_CONFIG_FILE_PATH',
    required=False,
    help='Filepath to PKS config file')
@click.option(
    '-s',
    '--skip-config-decryption',
    is_flag=True,
    help='Skip decryption of CSE/PKS config file')
@click.option(
    '-i',
    '--check-install',
    'check_install',
    is_flag=True,
    help='Checks that CSE/PKS is installed on vCD according '
         'to the config file')
def check(ctx, config_file_path, pks_config_file_path, skip_config_decryption,
          check_install):
    """Validate CSE config file."""
    SERVER_CLI_LOGGER.debug(f"Executing command: {ctx.command_path}")
    console_message_printer = utils.ConsoleMessagePrinter()
    utils.check_python_version(console_message_printer)

    password = None
    if not skip_config_decryption:
        password = os.getenv('CSE_CONFIG_PASSWORD') or utils.prompt_text(
            PASSWORD_FOR_CONFIG_DECRYPTION_MSG,
            color='green', hide_input=True)

    config_dict = None
    try:
        config_dict = get_validated_config(
            config_file_name=config_file_path,
            pks_config_file_name=pks_config_file_path,
            skip_config_decryption=skip_config_decryption,
            decryption_password=password,
            log_wire_file=SERVER_CLI_WIRELOG_FILEPATH,
            logger_debug=SERVER_CLI_LOGGER,
            msg_update_callback=console_message_printer)

        if check_install:
            try:
                configure_cse.check_cse_installation(
                    config_dict, msg_update_callback=console_message_printer)
            except Exception as err:
                msg = f"Error : {err}\nCSE installation is invalid"
                SERVER_CLI_LOGGER.error(msg)
                console_message_printer.error(msg)
                raise

        # Record telemetry data on successful completion
        record_user_action(
            cse_operation=CseOperation.CONFIG_CHECK,
            telemetry_settings=config_dict['service']['telemetry'])
        # Telemetry data construction
        cse_params = {
            PayloadKey.WAS_PKS_CONFIG_FILE_PROVIDED: bool(pks_config_file_path),  # noqa: E501
            PayloadKey.WAS_DECRYPTION_SKIPPED: bool(skip_config_decryption),
            PayloadKey.WAS_INSTALLATION_CHECKED: bool(check_install)
        }
        # Record telemetry data
        record_user_action_details(
            cse_operation=CseOperation.CONFIG_CHECK,
            cse_params=cse_params,
            telemetry_settings=config_dict['service']['telemetry'])
    except Exception as err:
        # Record telemetry data on failed operation
        if config_dict:
            record_user_action(
                cse_operation=CseOperation.CONFIG_CHECK,
                status=OperationStatus.FAILED,
                telemetry_settings=config_dict['service']['telemetry'])
        console_message_printer.error(str(err))
        SERVER_CLI_LOGGER.error(str(err))
        sys.exit(1)
    finally:
        # block the process to let telemetry handler to finish posting data to
        # VAC. HACK!!!
        time.sleep(3)


@cli.command(short_help='Decrypt the given file')
@click.pass_context
@click.argument('input_file', metavar='INPUT_FILE',
                type=click.Path(exists=True))
@click.option(
    '-o',
    '--output',
    'output_file',
    required=False,
    default=None,
    metavar='OUTPUT_FILE',
    help='Filepath to write decrypted file to')
def decrypt(ctx, input_file, output_file):
    """Decrypt CSE configuration file."""
    SERVER_CLI_LOGGER.debug(f"Executing command: {ctx.command_path}")
    console_message_printer = utils.ConsoleMessagePrinter()
    utils.check_python_version(console_message_printer)

    try:
        try:
            password = os.getenv('CSE_CONFIG_PASSWORD') or utils.prompt_text(
                PASSWORD_FOR_CONFIG_DECRYPTION_MSG,
                hide_input=True,
                color='green')
            decrypt_file(input_file, password, output_file)
            msg = "Decryption successful."
            console_message_printer.general(msg)
            SERVER_CLI_LOGGER.debug(msg)
        except cryptography.fernet.InvalidToken:
            raise Exception("Decryption failed: Invalid password")
    except Exception as err:
        console_message_printer.error(str(err))
        SERVER_CLI_LOGGER.error(str(err))
        sys.exit(1)


@cli.command(short_help='Encrypt the given file')
@click.pass_context
@click.argument('input_file', metavar='INPUT_FILE',
                type=click.Path(exists=True))
@click.option(
    '-o',
    '--output',
    'output_file',
    required=False,
    default=None,
    metavar='OUTPUT_FILE',
    help='Filepath to write encrypted file to')
def encrypt(ctx, input_file, output_file):
    """Encrypt CSE configuration file."""
    SERVER_CLI_LOGGER.debug(f"Executing command: {ctx.command_path}")
    console_message_printer = utils.ConsoleMessagePrinter()
    utils.check_python_version(console_message_printer)

    try:
        try:
            password = os.getenv('CSE_CONFIG_PASSWORD') or utils.prompt_text(
                PASSWORD_FOR_CONFIG_ENCRYPTION_MSG,
                hide_input=True,
                color='green')
            encrypt_file(input_file, password, output_file)
            msg = "Encryption successful."
            console_message_printer.general(msg)
            SERVER_CLI_LOGGER.debug(msg)
        except cryptography.fernet.InvalidToken:
            raise Exception("Encryption failed: Invalid password")
    except Exception as err:
        console_message_printer.error(str(err))
        SERVER_CLI_LOGGER.error(str(err))
        sys.exit(1)


@cli.command(short_help='Install CSE extension 3.1.0 on vCD')
@click.pass_context
@click.option(
    '-c',
    '--config',
    'config_file_path',
    default='config.yaml',
    metavar='CONFIG_FILE_PATH',
    type=click.Path(exists=True),
    envvar='CSE_CONFIG',
    required=True,
    help="(Required) Filepath to CSE config file")
@click.option(
    '-p',
    '--pks-config-file',
    'pks_config_file_path',
    type=click.Path(exists=True),
    metavar='PKS_CONFIG_FILE_PATH',
    required=False,
    help='Filepath to PKS config file')
@click.option(
    '-s',
    '--skip-config-decryption',
    is_flag=True,
    help='Skip decryption of CSE/PKS config file')
@click.option(
    '-t',
    '--skip-template-creation',
    'skip_template_creation',
    is_flag=True,
    help='Skips creating CSE k8s template during installation')
@click.option(
    '-d',
    '--retain-temp-vapp',
    'retain_temp_vapp',
    is_flag=True,
    help='Retain the temporary vApp after the template has been captured'
         ' --ssh-key option is required if this flag is used')
@click.option(
    '-k',
    '--ssh-key',
    'ssh_key_file',
    required=False,
    default=None,
    type=click.File('r'),
    help='Filepath of SSH public key to add to vApp template')
def install(ctx, config_file_path, pks_config_file_path,
            skip_config_decryption, skip_template_creation,
            retain_temp_vapp, ssh_key_file):
    """Install CSE on vCloud Director.

\b
    - legacy_mode is config property that is used to configure CSE with
      desired version of VCD.
    - Set legacy_mode=true if CSE 3.1 is configured with VCD whose maximum
      supported api_version < 35.
    - Set legacy_mode=false if CSE 3.1 is configured with VCD whose maximum
      supported api_version >= 35.
    - Note: legacy_mode=true is a valid condition for CSE 3.1 configured with
      VCD whose maximum supported api_version >= 35. However, it is strongly
      recommended to set the property to false to leverage the new
      functionality.
    - When legacy_mode=true, supported template information are based on
      remote-template-cookbook version "1.0.0".
    - When legacy_mode=false, supported template information are based on
      remote-template-cookbook version "2.0.0", min_cse_version and
      max_cse_version.
    """
    # NOTE: For CSE 3.0, if `enable_tkg_plus` in config file is set to false
    # and if `cse install` is invoked without skipping template creation,
    # an Exception will be thrown if TKG+ template is present in the
    # remote_template_cookbook.
    SERVER_CLI_LOGGER.debug(f"Executing command: {ctx.command_path}")
    console_message_printer = utils.ConsoleMessagePrinter()
    utils.check_python_version(console_message_printer)

    if retain_temp_vapp and not ssh_key_file:
        msg = "Must provide ssh-key file (using --ssh-key OR -k) if " \
              "--retain-temp-vapp is provided, or else temporary vm will be " \
              "inaccessible"
        console_message_printer.error(msg)
        SERVER_CLI_LOGGER.error(msg)
        sys.exit(1)

    ssh_key = None
    if ssh_key_file is not None:
        ssh_key = ssh_key_file.read()

    password = None
    if not skip_config_decryption:
        password = os.getenv('CSE_CONFIG_PASSWORD') or utils.prompt_text(
            PASSWORD_FOR_CONFIG_DECRYPTION_MSG,
            color='green', hide_input=True)

    try:
        config = get_validated_config(
            config_file_name=config_file_path,
            pks_config_file_name=pks_config_file_path,
            skip_config_decryption=skip_config_decryption,
            decryption_password=password,
            log_wire_file=INSTALL_WIRELOG_FILEPATH,
            logger_debug=INSTALL_LOGGER,
            msg_update_callback=console_message_printer)

        configure_cse.install_cse(
            config_file_name=config_file_path,
            config=config,
            pks_config_file_name=pks_config_file_path,
            skip_template_creation=skip_template_creation,
            ssh_key=ssh_key,
            retain_temp_vapp=retain_temp_vapp,
            skip_config_decryption=skip_config_decryption,
            msg_update_callback=console_message_printer)
    except Exception as err:
        SERVER_CLI_LOGGER.error(str(err))
        console_message_printer.error(str(err))
        sys.exit(1)
    finally:
        # block the process to let telemetry handler to finish posting data to
        # VAC. HACK!!!
        time.sleep(3)


@cli.command(short_help='Configure PKS')
@click.pass_context
@click.argument(
    'pks_config_file_path',
    type=click.Path(exists=True),
    metavar='PKS_CONFIG_FILE_PATH')
@click.option(
    '-s',
    '--skip-config-decryption',
    is_flag=True,
    help='Skip decryption of PKS config file')
def pks_configure(ctx, pks_config_file_path, skip_config_decryption):

    SERVER_CLI_LOGGER.debug(f"Executing command: {ctx.command_path}")
    console_message_printer = utils.ConsoleMessagePrinter()
    utils.check_python_version(console_message_printer)
    requests.packages.urllib3.disable_warnings()

    password = None
    if not skip_config_decryption:
        password = os.getenv('CSE_CONFIG_PASSWORD') or utils.prompt_text(
            PASSWORD_FOR_CONFIG_DECRYPTION_MSG,
            color='green', hide_input=True)
    try:
        try:
            utils.check_file_permissions(
                pks_config_file_path,
                msg_update_callback=console_message_printer)
            # TODO: reading pks config to be made common for cse install
            if skip_config_decryption:
                with open(pks_config_file_path) as f:
                    pks_config = yaml.safe_load(f) or {}
            else:
                console_message_printer.info(
                    f"Decrypting '{pks_config_file_path}'")
                pks_config = yaml.safe_load(
                    get_decrypted_file_contents(pks_config_file_path,
                                                password)) or {}
            configure_cse.configure_nsxt_for_cse(
                nsxt_servers=pks_config['nsxt_servers'],
                log_wire=True,
                msg_update_callback=console_message_printer
            )
        except requests.exceptions.SSLError as err:
            raise Exception(f"SSL verification failed: {str(err)}")
        except requests.exceptions.ConnectionError as err:
            raise Exception(f"Cannot connect to {err.request.url}.")
        except cryptography.fernet.InvalidToken:
            raise Exception(CONFIG_DECRYPTION_ERROR_MSG)
    except Exception as err:
        SERVER_CLI_LOGGER.error(str(err))
        console_message_printer.error(str(err))
        sys.exit(1)


@cli.command(short_help='Run CSE service')
@click.pass_context
@click.option(
    '-c',
    '--config',
    'config_file_path',
    default='config.yaml',
    metavar='CONFIG_FILE_PATH',
    type=click.Path(exists=True),
    envvar='CSE_CONFIG',
    required=True,
    help="(Required) Filepath to CSE config file")
@click.option(
    '-p',
    '--pks-config-file',
    'pks_config_file_path',
    type=click.Path(exists=True),
    metavar='PKS_CONFIG_FILE_PATH',
    required=False,
    help='Filepath to PKS config file')
@click.option(
    '--skip-check',
    is_flag=True,
    help='Skip CSE installation checks')
@click.option(
    '-s',
    '--skip-config-decryption',
    is_flag=True,
    help='Skip decryption of CSE/PKS config file')
def run(ctx, config_file_path, pks_config_file_path, skip_check,
        skip_config_decryption):
    """Run CSE service.

    legacy_mode is config property that is used to configure CSE with
    desired version of VCD.
    Set legacy_mode=true if CSE 3.1 is configured with VCD whose maximum
    supported api_version < 35.
    Set legacy_mode=false if CSE 3.1 is configured with VCD whose maximum
    supported api_version >= 35.
    Note: legacy_mode=true is a valid condition for CSE 3.1 configured with
    VCD whose maximum supported api_version >= 35. However, it is strongly
    recommended to set the property to false to leverage the new functionality.

    """
    SERVER_CLI_LOGGER.debug(f"Executing command: {ctx.command_path}")
    console_message_printer = utils.ConsoleMessagePrinter()
    utils.check_python_version(console_message_printer)

    password = None
    if not skip_config_decryption:
        password = os.getenv('CSE_CONFIG_PASSWORD') or utils.prompt_text(
            PASSWORD_FOR_CONFIG_DECRYPTION_MSG,
            color='green', hide_input=True)

    config = None
    cse_run_complete = False
    try:
        config = get_validated_config(
            config_file_name=config_file_path,
            pks_config_file_name=pks_config_file_path,
            skip_config_decryption=skip_config_decryption,
            decryption_password=password,
            log_wire_file=SERVER_DEBUG_WIRELOG_FILEPATH,
            logger_debug=SERVER_LOGGER,
            msg_update_callback=console_message_printer)
        service = cse_service.Service(
            config_file=config_file_path,
            config=config,
            pks_config_file=pks_config_file_path,
            should_check_config=not skip_check,
            skip_config_decryption=skip_config_decryption)
        service.run(msg_update_callback=console_message_printer)
        cse_run_complete = True
    except Exception as err:
        SERVER_CLI_LOGGER.error(str(err))
        console_message_printer.error(str(err))
        console_message_printer.error("CSE Server failure. Please check the logs.")  # noqa: E501
        sys.exit(1)
    finally:
        if not cse_run_complete:
            telemetry_settings = config['service']['telemetry'] if config else None  # noqa: E501
            record_user_action(cse_operation=CseOperation.SERVICE_RUN,
                               status=OperationStatus.FAILED,
                               telemetry_settings=telemetry_settings)  # noqa: E501
            # block the process to let telemetry handler to finish posting
            # data to VAC. HACK!!!
            time.sleep(3)


@cli.command('upgrade',
             short_help="Upgrade CSE extension to version 3.1.0 on vCD")
@click.pass_context
@click.option(
    '-c',
    '--config',
    'config_file_path',
    default='config.yaml',
    metavar='CONFIG_FILE_PATH',
    type=click.Path(exists=True),
    envvar='CSE_CONFIG',
    required=True,
    help="Filepath to CSE config file")
@click.option(
    '-s',
    '--skip-config-decryption',
    'skip_config_decryption',
    is_flag=True,
    help='Skip decryption of CSE config file')
@click.option(
    '-t',
    '--skip-template-creation',
    'skip_template_creation',
    is_flag=True,
    help='Skip creating CSE k8s templates during upgrade')
@click.option(
    '-d',
    '--retain-temp-vapp',
    'retain_temp_vapp',
    is_flag=True,
    help='Retain the temporary vApp after the CSE k8s template has been captured'  # noqa: E501
         ' --ssh-key option is required if this flag is used')
@click.option(
    '-k',
    '--ssh-key',
    'ssh_key_file',
    required=False,
    default=None,
    type=click.File('r'),
    help='Filepath of SSH public key to add to CSE k8s template vms')
def upgrade(ctx, config_file_path, skip_config_decryption,
            skip_template_creation, retain_temp_vapp,
            ssh_key_file):
    """Upgrade existing CSE installation/entities to match CSE 3.1.0.

\b
    - Add CSE, RDE version, Legacy mode info to VCD's extension data for CSE
    - Register defined entities schema of CSE k8s clusters with VCD
    - Create placement compute policies used by CSE
    - Remove old sizing compute policies created by CSE 2.6 and below
    - Install all templates from template repository linked in config file
    - Currently installed templates that are no longer compliant with
      new CSE template cookbook will not be recognized by CSE 3.1. Admins can
      safely delete them once new templates are installed and the existing
      clusters are upgraded to newer template revisions.
    - Update existing CSE k8s cluster's to match CSE 3.1 k8s clusters.
    - Upgrading legacy clusters would require new template creation supported
      by CSE 3.1.0
    - legacy_mode is config property that is used to configure CSE with
      desired version of VCD.
    - Set legacy_mode=true if CSE 3.1 is configured with VCD whose maximum
      supported api_version < 35.
    - Set legacy_mode=false if CSE 3.1 is configured with VCD whose maximum
      supported api_version >= 35.
    - NOTE: legacy_mode=true is a valid condition for CSE 3.1 configured with
      VCD whose maximum supported api_version >= 35. However, it is strongly
      recommended to set the property to false to leverage
      the new functionality.
    - When legacy_mode=true, supported template information are based on
      remote-template-cookbook version "1.0.0".
    - When legacy_mode=false, supported template information are based on
      remote-template-cookbook version "2.0.0", min_cse_version and
      max_cse_version.
    """
    # NOTE: For CSE 3.0, if `enable_tkg_plus` in the config is set to false,
    # an exception is thrown if
    # 1. If there is an existing TKG+ template
    # 2. If remote template cookbook contains a TKG+ template.
    SERVER_CLI_LOGGER.debug(f"Executing command: {ctx.command_path}")
    console_message_printer = utils.ConsoleMessagePrinter()
    utils.check_python_version(console_message_printer)

    if retain_temp_vapp and not ssh_key_file:
        msg = "Must provide ssh-key file (using --ssh-key OR -k) if " \
              "--retain-temp-vapp is provided, or else temporary vm will be " \
              "inaccessible"
        console_message_printer.error(msg)
        INSTALL_LOGGER.error(msg)
        sys.exit(1)

    ssh_key = None
    if ssh_key_file is not None:
        ssh_key = ssh_key_file.read()

    password = None
    if not skip_config_decryption:
        password = os.getenv('CSE_CONFIG_PASSWORD') or utils.prompt_text(
            PASSWORD_FOR_CONFIG_DECRYPTION_MSG,
            color='green', hide_input=True)

    try:
        config = get_validated_config(
            config_file_name=config_file_path,
            pks_config_file_name='',
            skip_config_decryption=skip_config_decryption,
            decryption_password=password,
            log_wire_file=INSTALL_WIRELOG_FILEPATH,
            logger_debug=INSTALL_LOGGER,
            msg_update_callback=console_message_printer)

        configure_cse.upgrade_cse(
            config_file_name=config_file_path,
            config=config,
            skip_template_creation=skip_template_creation,
            ssh_key=ssh_key,
            retain_temp_vapp=retain_temp_vapp,
            msg_update_callback=console_message_printer)

    except Exception as err:
        SERVER_CLI_LOGGER.error(str(err))
        console_message_printer.error(str(err))
        sys.exit(1)
    finally:
        # block the process to let telemetry handler to finish posting data to
        # VAC. HACK!!!
        time.sleep(3)


@template.command(
    'list',
    short_help='List native Kubernetes templates')
@click.pass_context
@click.option(
    '-c',
    '--config',
    'config_file_path',
    default='config.yaml',
    metavar='CONFIG_FILE_PATH',
    type=click.Path(exists=True),
    envvar='CSE_CONFIG',
    required=True,
    help="(Required) Filepath to CSE config file")
@click.option(
    '-s',
    '--skip-config-decryption',
    is_flag=True,
    help='Skip decryption of CSE config file')
@click.option(
    '-d',
    '--display',
    'display_option',
    type=click.Choice(
        [DISPLAY_ALL, DISPLAY_DIFF, DISPLAY_LOCAL, DISPLAY_REMOTE]),
    default=DISPLAY_ALL,
    help='Choose templates to display.')
def list_template(ctx, config_file_path, skip_config_decryption,
                  display_option):
    """List CSE k8s templates."""
    # TODO(VCDA-2236) Validate legacy_mode in extension
    SERVER_CLI_LOGGER.debug(f"Executing command: {ctx.command_path}")
    console_message_printer = utils.ConsoleMessagePrinter()
    # Not passing the console_message_printer, because we want to suppress
    # the python version check messages from being printed onto console.
    utils.check_python_version()

    config_dict = None
    try:
        # We don't want to validate config file, because server startup or
        # installation is not being performed. If values in config file are
        # missing or bad, appropriate exception will be raised while accessing
        # or using them.
        config_dict = _get_unvalidated_config(
            config_file_path=config_file_path,
            skip_config_decryption=skip_config_decryption,
            msg_update_callback=console_message_printer)

        # Record telemetry details
        cse_params = {
            PayloadKey.DISPLAY_OPTION: display_option,
            PayloadKey.WAS_DECRYPTION_SKIPPED: bool(skip_config_decryption)
        }
        record_user_action_details(cse_operation=CseOperation.TEMPLATE_LIST,
                                   cse_params=cse_params,
                                   telemetry_settings=config_dict['service']['telemetry'])  # noqa: E501

        local_templates = []
        if display_option in (DISPLAY_ALL, DISPLAY_DIFF, DISPLAY_LOCAL):
            client = None
            try:
                log_wire_file = None
                log_wire = utils.str_to_bool(config_dict['service'].get('log_wire'))  # noqa: E501
                if log_wire:
                    log_wire_file = SERVER_DEBUG_WIRELOG_FILEPATH

                # Note: This will get us a client with highest supported
                # VCD api version.
                client, _ = _get_clients_from_config(config_dict,
                                                     log_wire_file=log_wire_file,  # noqa: E501
                                                     log_wire=log_wire)

                org_name = config_dict['broker']['org']
                catalog_name = config_dict['broker']['catalog']
                is_tkg_plus_enabled = server_utils.is_tkg_plus_enabled(config_dict)  # noqa: E501
                legacy_mode = config_dict['service']['legacy_mode']

                local_template_definitions = \
                    ltm.get_valid_k8s_local_template_definition(
                        client=client,
                        catalog_name=catalog_name,
                        legacy_mode=legacy_mode,
                        is_tkg_plus_enabled=is_tkg_plus_enabled,
                        org_name=org_name,
                        logger_debug=SERVER_CLI_LOGGER)
                default_template_name = \
                    config_dict['broker']['default_template_name']
                default_template_revision = \
                    str(config_dict['broker']['default_template_revision'])

                for definition in local_template_definitions:
                    local_template = {
                        'name': definition[LocalTemplateKey.NAME],
                        'revision': int(definition[LocalTemplateKey.REVISION]),
                        'local': 'Yes',
                        'remote': 'No',
                        'cpu': definition[LocalTemplateKey.CPU],
                        'memory': definition[LocalTemplateKey.MEMORY],
                        'description': definition[LocalTemplateKey.DESCRIPTION]
                    }
                    if legacy_mode:
                        local_template['compute_policy'] = \
                            definition[LegacyLocalTemplateKey.COMPUTE_POLICY]
                    # Any metadata read from vCD is sting due to how pyvcloud
                    # is coded, so we need to cast it back to int.
                    if (definition[LocalTemplateKey.NAME], str(definition[LocalTemplateKey.REVISION])) == (default_template_name, default_template_revision):  # noqa: E501
                        local_template['default'] = 'Yes'
                    else:
                        local_template['default'] = 'No'
                    local_template['deprecated'] = 'Yes' if utils.str_to_bool(definition[LocalTemplateKey.DEPRECATED]) else 'No'  # noqa: E501

                    local_templates.append(local_template)
            finally:
                if client:
                    client.logout()

        remote_templates = []
        if display_option in (DISPLAY_ALL, DISPLAY_DIFF, DISPLAY_REMOTE):
            rtm = RemoteTemplateManager(
                remote_template_cookbook_url=config_dict['broker']['remote_template_cookbook_url'],  # noqa: E501
                legacy_mode=config_dict['service']['legacy_mode'],
                logger=SERVER_CLI_LOGGER)
            remote_template_cookbook = \
                rtm.get_filtered_remote_template_cookbook()
            remote_template_definitions = remote_template_cookbook['templates']
            remote_template_keys = server_utils.get_template_descriptor_keys(
                rtm.cookbook_version)
            for definition in remote_template_definitions:
                remote_template = {
                    'name': definition[remote_template_keys.NAME],
                    'revision': definition[remote_template_keys.REVISION],
                    'local': 'No',
                    'remote': 'Yes',
                    'cpu': definition[remote_template_keys.CPU],
                    'memory': definition[remote_template_keys.MEMORY],
                    'description': definition[remote_template_keys.DESCRIPTION]
                }
                if legacy_mode:
                    remote_template['compute_policy'] = \
                        definition[remote_template_keys.COMPUTE_POLICY]
                if display_option is DISPLAY_ALL:
                    remote_template['default'] = 'No'
                remote_template['deprecated'] = 'Yes' if utils.str_to_bool(definition[remote_template_keys.DEPRECATED]) else 'No'  # noqa: E501

                remote_templates.append(remote_template)

        result = []
        if display_option is DISPLAY_ALL:
            result = remote_templates
            # If local copy of template exists, update the remote definition
            # with relevant values, else add the local definition to the result
            # list.
            for local_template in local_templates:
                found = False
                for remote_template in remote_templates:
                    if (local_template[LocalTemplateKey.NAME], local_template[LocalTemplateKey.REVISION]) == (remote_template[remote_template_keys.NAME], remote_template[remote_template_keys.REVISION]):  # noqa: E501
                        if legacy_mode:
                            remote_template['compute_policy'] = \
                                local_template['compute_policy']
                        remote_template['local'] = local_template['local']
                        remote_template['default'] = local_template['default']
                        found = True
                        break
                if not found:
                    result.append(local_template)
        elif display_option in DISPLAY_DIFF:
            for remote_template in remote_templates:
                found = False
                for local_template in local_templates:
                    if (local_template[LocalTemplateKey.NAME], local_template[LocalTemplateKey.REVISION]) == (remote_template[remote_template_keys.NAME], remote_template[remote_template_keys.REVISION]):  # noqa: E501
                        found = True
                        break
                if not found:
                    result.append(remote_template)
        elif display_option in DISPLAY_LOCAL:
            result = local_templates
        elif display_option in DISPLAY_REMOTE:
            result = remote_templates

        result = sorted(result, key=lambda t: (t['name'], t['revision']), reverse=True)  # noqa: E501
        stdout(result, ctx, sort_headers=False)
        SERVER_CLI_LOGGER.debug(result)
        record_user_action(cse_operation=CseOperation.TEMPLATE_LIST,
                           telemetry_settings=config_dict['service']['telemetry'])  # noqa: E501
    except Exception as err:
        SERVER_CLI_LOGGER.error(str(err))
        console_message_printer.error(str(err))
        telemetry_settings = config_dict.get('service', {}).get('telemetry') \
            if config_dict else None
        record_user_action(cse_operation=CseOperation.TEMPLATE_LIST,
                           status=OperationStatus.FAILED,
                           telemetry_settings=telemetry_settings)
        sys.exit(1)
    finally:
        # block the process to let telemetry handler to finish posting data to
        # VAC. HACK!!!
        time.sleep(3)


@template.command('install',
                  short_help="Create native Kubernetes templates listed in "
                             "remote template repository. Use '*' for "
                             "TEMPLATE_NAME and TEMPLATE_REVISION to install "
                             "all listed templates.")
@click.pass_context
@click.argument('template_name', metavar='TEMPLATE_NAME', default='*')
@click.argument('template_revision', metavar='TEMPLATE_REVISION', default='*')
@click.option(
    '-c',
    '--config',
    'config_file_path',
    default='config.yaml',
    metavar='CONFIG_FILE_PATH',
    type=click.Path(exists=True),
    envvar='CSE_CONFIG',
    required=True,
    help="(Required) Filepath to CSE config file")
@click.option(
    '-s',
    '--skip-config-decryption',
    is_flag=True,
    help='Skip decryption of CSE config file')
@click.option(
    '-f',
    '--force',
    'force_create',
    is_flag=True,
    help='Recreate CSE k8s templates on vCD even if they already exist')
@click.option(
    '-d',
    '--retain-temp-vapp',
    'retain_temp_vapp',
    is_flag=True,
    help='Retain the temporary vApp after the template has been captured'
         ' --ssh-key option is required if this flag is used')
@click.option(
    '-k',
    '--ssh-key',
    'ssh_key_file',
    required=False,
    default=None,
    type=click.File('r'),
    help='Filepath of SSH public key to add to vApp template')
def install_cse_template(ctx, template_name, template_revision,
                         config_file_path, skip_config_decryption,
                         force_create, retain_temp_vapp,
                         ssh_key_file):
    """Create Kubernetes templates listed in remote template repository.

    Use '*' for TEMPLATE_NAME and TEMPLATE_REVISION to install
    all listed templates.

    legacy_mode is config property that is used to configure CSE with
    desired version of VCD.

    When legacy_mode=true, supported template information are based on
    remote-template-cookbook version "1.0.0".

    When legacy_mode=false, supported template information are based on
    remote-template-cookbook version "2.0.0", min_cse_version and
    max_cse_version.
    """
    # NOTE: For CSE 3.0, if `enable_tkg_plus` flag in config is set to false,
    # Throw an error if TKG+ template creation is issued.
    SERVER_CLI_LOGGER.debug(f"Executing command: {ctx.command_path}")
    console_message_printer = utils.ConsoleMessagePrinter()
    utils.check_python_version(console_message_printer)

    if retain_temp_vapp and not ssh_key_file:
        msg = "Must provide ssh-key file (using --ssh-key OR -k) if " \
              "--retain-temp-vapp is provided, or else temporary vm will be " \
              "inaccessible"
        SERVER_CLI_LOGGER.error(msg)
        console_message_printer.error(msg)
        sys.exit(1)

    password = None
    if not skip_config_decryption:
        password = os.getenv('CSE_CONFIG_PASSWORD') or utils.prompt_text(
            PASSWORD_FOR_CONFIG_DECRYPTION_MSG,
            color='green', hide_input=True)

    ssh_key = None
    if ssh_key_file is not None:
        ssh_key = ssh_key_file.read()

    try:
        config = get_validated_config(
            config_file_name=config_file_path,
            skip_config_decryption=skip_config_decryption,
            decryption_password=password,
            log_wire_file=INSTALL_WIRELOG_FILEPATH,
            logger_debug=INSTALL_LOGGER,
            msg_update_callback=console_message_printer)

        configure_cse.install_template(
            template_name=template_name,
            template_revision=template_revision,
            config_file_name=config_file_path,
            config=config,
            force_create=force_create,
            retain_temp_vapp=retain_temp_vapp,
            ssh_key=ssh_key,
            skip_config_decryption=skip_config_decryption,
            msg_update_callback=console_message_printer)
    except Exception as err:
        SERVER_CLI_LOGGER.error(str(err))
        console_message_printer.error(str(err))
        sys.exit(1)
    finally:
        # block the process to let telemetry handler to finish posting data to
        # VAC. HACK!!!
        time.sleep(3)


def _get_unvalidated_config(config_file_path,
                            skip_config_decryption,
                            msg_update_callback=utils.NullPrinter()):
    password = None
    if not skip_config_decryption:
        password = os.getenv('CSE_CONFIG_PASSWORD') or utils.prompt_text(
            PASSWORD_FOR_CONFIG_DECRYPTION_MSG,
            color='green', hide_input=True)

    try:
        if skip_config_decryption:
            with open(config_file_path) as config_file:
                config_dict = yaml.safe_load(config_file) or {}
        else:
            msg_update_callback.info(
                f"Decrypting '{config_file_path}'")
            config_dict = yaml.safe_load(
                get_decrypted_file_contents(
                    config_file_path, password)) or {}
        msg_update_callback.general(f"Retrieved config from "
                                    f"'{config_file_path}'")

        # To suppress the warning message that pyvcloud prints if
        # ssl_cert verification is skipped.
        if config_dict and config_dict.get('vcd') and \
                not config_dict['vcd'].get('verify'):
            requests.packages.urllib3.disable_warnings()

        # Store telemetry instance id, url and collector id in config
        # This step should be done after suppressing the cert validation
        # warnings
        store_telemetry_settings(config_dict)

        return config_dict
    except cryptography.fernet.InvalidToken:
        raise Exception(CONFIG_DECRYPTION_ERROR_MSG)


def _get_clients_from_config(config, log_wire_file, log_wire):
    client = vcd_client.Client(
        config['vcd']['host'],
        verify_ssl_certs=config['vcd']['verify'],
        log_file=log_wire_file,
        log_requests=log_wire,
        log_headers=log_wire,
        log_bodies=log_wire)
    credentials = vcd_client.BasicLoginCredentials(
        config['vcd']['username'],
        SYSTEM_ORG_NAME,
        config['vcd']['password'])
    client.set_credentials(credentials)

    logger_wire = NULL_LOGGER
    if log_wire:
        logger_wire = SERVER_CLOUDAPI_WIRE_LOGGER

    cloudapi_client = vcd_utils.get_cloudapi_client_from_vcd_client(
        client,
        SERVER_CLI_LOGGER,
        logger_wire)

    return client, cloudapi_client


if __name__ == '__main__':
    cli()



# container-service-extension
# Copyright (c) 2021 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

from dataclasses import dataclass
from typing import List, Optional

from dataclasses_json import dataclass_json, LetterCase
import yaml

import container_service_extension.common.constants.server_constants as server_constants  # noqa: E501
import container_service_extension.common.constants.shared_constants as shared_constants  # noqa: E501
import container_service_extension.rde.constants as rde_constants
from container_service_extension.rde.models.abstractNativeEntity import AbstractNativeEntity  # noqa: E501
import container_service_extension.rde.models.rde_1_0_0 as rde_1_0_0


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Metadata:
    name: str
    org_name: str
    virtual_data_center_name: str
    site: str


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Nfs:
    sizing_class: Optional[str] = None
    storage_profile: Optional[str] = None
    count: int = 0


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ControlPlane:
    sizing_class: Optional[str] = None
    storage_profile: Optional[str] = None
    count: int = 1
    # cpu in Ghz
    cpu: Optional[int] = None
    # memory in Mb
    memory: Optional[int] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Workers:
    sizing_class: Optional[str] = None
    storage_profile: Optional[str] = None
    count: int = 0
    # cpu in Ghz
    cpu: Optional[int] = None
    # memory in Mb
    memory: Optional[int] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Distribution:
    template_name: Optional[str] = ''
    template_revision: Optional[int] = 0


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Topology:
    control_plane: ControlPlane = ControlPlane()
    workers: Workers = Workers()
    nfs: Nfs = Nfs()


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Cni:
    name: Optional[str] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Pods:
    cidr_blocks: Optional[List[str]] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Services:
    cidr_blocks: Optional[List[str]] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Network:
    cni: Optional[Cni] = None
    pods: Optional[Pods] = None
    services: Optional[Services] = None
    expose: bool = False


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Settings:
    ovdc_network: str
    ssh_key: Optional[str] = None
    rollback_on_failure: bool = True
    network: Network = Network()


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Node:
    name: str
    ip: str
    sizing_class: Optional[str] = None
    storage_profile: Optional[str] = None
    # cpu in Ghz
    cpu: Optional[int] = None
    # memory in Mb
    memory: Optional[int] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class NfsNode(Node):
    exports: Optional[List[str]] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Nodes:
    control_plane: Optional[Node] = None
    workers: Optional[List[Node]] = None
    nfs: Optional[List[NfsNode]] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class CloudProperties:
    site: Optional[str] = None
    org_name: Optional[str] = None
    virtual_data_center_name: Optional[str] = None
    ovdc_network_name: Optional[str] = None
    distribution: Distribution = Distribution()

    # TODO contemplate the need to keep this attribute
    ssh_key: Optional[str] = None

    rollback_on_failure: bool = True
    exposed: bool = False


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Private:
    kube_config: Optional[str] = None
    kube_token: Optional[str] = None
    certificates: Optional[List[str]] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Status:
    phase: Optional[str] = None
    cni: Optional[str] = None
    task_href: Optional[str] = None
    kubernetes: Optional[str] = None
    docker_version: Optional[str] = None
    os: Optional[str] = None
    external_ip: Optional[str] = None
    nodes: Optional[Nodes] = None
    uid: Optional[str] = None
    cloud_properties: CloudProperties = CloudProperties()
    persistent_volumes: Optional[List[str]] = None
    virtual_IPs: Optional[List[str]] = None
    private: Optional[Private] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ClusterSpec:
    """Represents the cluster spec.

    If dictionaries are passed as arguments, the constructor auto-converts
    them into the expected class instances.
    """

    settings: Settings
    topology: Topology = Topology()
    distribution: Distribution = Distribution()


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class NativeEntity(AbstractNativeEntity):
    # TODO change the comment to have camelCase
    """Represents the Native Cluster entity.

    If dictionaries are passed as arguments, the constructor auto-converts
    them into the expected class instances.

    Sample representation in JSON format
    # TODO change sample representation
    {
        "kind": "native",
        "spec": {
            "topology": {
                "workers": {
                    "count": 2,
                    "sizingClass": "small",
                    "storageProfile": "Any"
                },
                "controlPlane": {
                    "count": 1,
                    "sizingClass": "Large",
                    "storageProfile": "Any"
                },
            }
            "settings": {
                "ovdc_network": "net",
                "sshKey": null,
                "rollbackOnFailure": true
                "network" : {
                    "expose" : false
                }
            },
            "distribution": {
                "templateName": "k81.17",
                "templateRevision": 1
            }
        },
        "status": {
            "id": null,
            "cni": null,
            "phase": null,
        },
        "metadata": {
            "orgName": "org1",
            "virtualDataCenterName": "ovdc1",
            "name": "myCluster"
            "site": "vcd_site"
        },
        "apiVersion": ""
    }
    """

    metadata: Metadata
    spec: ClusterSpec
    api_version: str
    status: Status = Status()
    kind: str = shared_constants.ClusterEntityKind.NATIVE.value

    @classmethod
    def from_native_entity(cls, native_entity: AbstractNativeEntity):
        """Construct rde_2.0.0 native entity from different rde_x.x.x.

        Use case: converts the NativeEntity of the specified RDE version
        to RDE 2.0.0.

        :param AbstractNativeEntity native_entity: input native entity
        :return: native entity
        :rtype: rde_2.0.0.NativeEntity
        """
        if isinstance(native_entity, NativeEntity):
            return native_entity

        if isinstance(native_entity, rde_1_0_0.NativeEntity):
            # TODO should change very much
            rde_1_x_entity: rde_1_0_0.NativeEntity = native_entity

            distribution = Distribution(
                template_name=rde_1_x_entity.spec.k8_distribution.template_name,  # noqa: E501
                template_revision=rde_1_x_entity.spec.k8_distribution.template_revision  # noqa: E501
            )
            # NOTE: since details for the field `site` is not present in
            # RDE 1.0, it is left empty
            cloud_properties = CloudProperties(
                site=None,
                org_name=rde_1_x_entity.metadata.org_name,
                virtual_data_center_name=rde_1_x_entity.metadata.ovdc_name,
                ovdc_network_name=rde_1_x_entity.spec.settings.network,
                distribution=distribution,
                ssh_key=rde_1_x_entity.spec.settings.ssh_key,
                rollback_on_failure=rde_1_x_entity.spec.settings.rollback_on_failure,  # noqa: E501
                exposed=rde_1_x_entity.status.exposed)
            # RDE 1.0 don't have storage_profile in Node definition
            # Convert "status" only for entities that are already created
            # New cluster creation won't have "status" section in RDE
            status = cls.status
            if rde_1_x_entity.status.nodes:
                control_plane = Node(
                    name=rde_1_x_entity.status.nodes.control_plane.name,
                    ip=rde_1_x_entity.status.nodes.control_plane.ip,
                    sizing_class=rde_1_x_entity.status.nodes.control_plane.sizing_class,  # noqa: E501
                    storage_profile=rde_1_x_entity.spec.control_plane.storage_profile  # noqa: E501
                )
                workers = []
                for worker in rde_1_x_entity.status.nodes.workers:
                    worker_node_2_x = Node(
                        name=worker.name,
                        ip=worker.ip,
                        sizing_class=worker.sizing_class,
                        storage_profile=rde_1_x_entity.spec.workers.storage_profile  # noqa: E501
                    )
                    workers.append(worker_node_2_x)
                nfs_nodes = []
                for nfs_node in rde_1_x_entity.status.nodes.nfs:
                    # The nfs_node.export field is a string
                    # however when it was created by cluster_service_1_x.py
                    # it just took a list and string-ified it. The piece of
                    # code below reverses the string representation of the list
                    # back into a list of strings.
                    if isinstance(nfs_node.exports, str):
                        export_list_string = nfs_node.exports
                        export_list_string.replace('[', '').replace(']', '').replace('\'', '')  # noqa: E501
                        export_list = export_list_string.split(", ")
                    elif isinstance(nfs_node.exports, list):
                        export_list = nfs_node.exports
                    else:
                        export_list = [str(nfs_node.exports)]

                    nfs_node_2_x = NfsNode(
                        name=nfs_node.name,
                        ip=nfs_node.ip,
                        sizing_class=nfs_node.sizing_class,
                        storage_profile=rde_1_x_entity.spec.nfs.storage_profile,  # noqa: E501
                        exports=export_list
                    )
                    nfs_nodes.append(nfs_node_2_x)
                nodes = Nodes(
                    control_plane=control_plane,
                    workers=workers,
                    nfs=nfs_nodes
                )

                external_ip = None
                if rde_1_x_entity.status.exposed:
                    external_ip = rde_1_x_entity.status.nodes.control_plane.ip

                # NOTE: since details for the field `uid` is not present in
                # RDE 1.0, it is left empty.
                # Proper value for `uid` should be populated after RDE is converted  # noqa: E501
                # as `uid` is a required property in Status for RDE 2.0
                status = Status(phase=rde_1_x_entity.status.phase,
                                cni=rde_1_x_entity.status.cni,
                                task_href=rde_1_x_entity.status.task_href,
                                kubernetes=rde_1_x_entity.status.kubernetes,
                                docker_version=rde_1_x_entity.status.docker_version,  # noqa: E501
                                os=rde_1_x_entity.status.os,
                                external_ip=external_ip,
                                nodes=nodes,
                                uid=None,
                                cloud_properties=cloud_properties)
            # NOTE: since details for the field `site` is not present in
            # RDE 1.0, it is left empty.
            # Proper value for `site` should be populated after RDE is
            # converted. Since, `site` is a required property in Metadata
            # for RDE 2.0
            metadata = Metadata(name=rde_1_x_entity.metadata.cluster_name,
                                org_name=rde_1_x_entity.metadata.org_name,
                                virtual_data_center_name=rde_1_x_entity.metadata.ovdc_name,  # noqa: E501
                                site='')
            topology = Topology(
                control_plane=ControlPlane(
                    sizing_class=rde_1_x_entity.spec.control_plane.sizing_class,  # noqa: E501
                    storage_profile=rde_1_x_entity.spec.control_plane.storage_profile,  # noqa: E501
                    count=rde_1_x_entity.spec.control_plane.count
                ),
                workers=Workers(
                    sizing_class=rde_1_x_entity.spec.workers.sizing_class,
                    storage_profile=rde_1_x_entity.spec.workers.storage_profile,  # noqa: E501
                    count=rde_1_x_entity.spec.workers.count
                ),
                nfs=Nfs(
                    sizing_class=rde_1_x_entity.spec.nfs.sizing_class,
                    storage_profile=rde_1_x_entity.spec.nfs.storage_profile,
                    count=rde_1_x_entity.spec.nfs.count
                ),
            )
            spec = ClusterSpec(
                settings=Settings(
                    ovdc_network=rde_1_x_entity.spec.settings.network,
                    ssh_key=rde_1_x_entity.spec.settings.ssh_key,
                    rollback_on_failure=rde_1_x_entity.spec.settings.rollback_on_failure,  # noqa: E501
                    network=Network(
                        expose=rde_1_x_entity.spec.expose
                    )
                ),
                topology=topology,
                distribution=Distribution(
                    template_name=rde_1_x_entity.spec.k8_distribution.template_name,  # noqa: E501
                    template_revision=rde_1_x_entity.spec.k8_distribution.template_revision  # noqa: E501
                )
            )
            rde_2_entity = cls(
                metadata=metadata,
                spec=spec,
                status=status,
                kind=rde_1_x_entity.kind,
                api_version=rde_constants.PAYLOAD_VERSION_2_0
            )
            return rde_2_entity

    @classmethod
    def from_cluster_data(cls, cluster: dict, kind: str, **kwargs):
        """Construct rde_2.0.0 native entity from non-rde cluster.

        :param dict cluster: cluster metadata
        :param str kind: cluster kind
        :return: native entity
        :rtype: rde_2.0.0.NativeEntity
        """
        site = kwargs.get('site', '')
        worker_nodes = []
        for item in cluster['nodes']:
            worker_nodes.append(
                Node(
                    name=item['name'],
                    ip=item['ipAddress'],
                    storage_profile=cluster['storage_profile_name']
                )
            )
        nfs_nodes = []
        for item in cluster['nfs_nodes']:
            # The item['exports'] field is a string
            # however when it was created by vcd_broker.py
            # it just took a list and string-ified it. The piece of
            # code below reverses the string representation of the list
            # back into a list of strings.
            exports_list_string = item['exports']
            exports_list_string.replace('[', '').replace(']', '').replace('\'', '')  # noqa: E501
            exports_list = exports_list_string.split(", ")

            nfs_nodes.append(
                NfsNode(
                    name=item['name'],
                    ip=item['ipAddress'],
                    storage_profile=cluster['storage_profile_name'],
                    exports=exports_list
                )
            )

        k8_distribution = Distribution(
            template_name=cluster['template_name'],
            template_revision=int(cluster['template_revision']))

        cloud_properties = CloudProperties(
            site=site,
            org_name=cluster['org_name'],
            virtual_data_center_name=cluster['vdc_name'],
            ovdc_network_name=cluster['network_name'],
            distribution=k8_distribution,
            ssh_key='')
        topology = Topology(
            workers=Workers(
                count=len(cluster['nodes']),
                storage_profile=cluster['storage_profile_name']
            ),
            control_plane=ControlPlane(
                count=len(cluster['master_nodes']),
                storage_profile=cluster['storage_profile_name']
            ),
            nfs=Nfs(
                count=len(cluster['nfs_nodes']),
                storage_profile=cluster['storage_profile_name']
            )
        )
        cluster_entity = cls(
            kind=kind,
            spec=ClusterSpec(
                topology=topology,
                settings=Settings(
                    ovdc_network=cluster['network_name'],
                    ssh_key=""
                ),
                distribution=k8_distribution
            ),
            status=Status(
                phase=str(server_constants.DefEntityPhase(
                    server_constants.DefEntityOperation.CREATE,
                    server_constants.DefEntityOperationStatus.SUCCEEDED)
                ),
                kubernetes=f"{cluster['kubernetes']} {cluster['kubernetes_version']}",  # noqa: E501
                cni=f"{cluster['cni']} {cluster['cni_version']}",
                os=cluster['os'],
                docker_version=cluster['docker_version'],
                nodes=Nodes(
                    control_plane=Node(
                        name=cluster['master_nodes'][0]['name'],
                        ip=cluster['master_nodes'][0]['ipAddress'],
                        storage_profile=cluster['storage_profile_name']
                    ),
                    workers=worker_nodes,
                    nfs=nfs_nodes
                ),
                uid=cluster['cluster_id'],
                cloud_properties=cloud_properties
            ),
            metadata=Metadata(
                site=site,
                org_name=cluster['org_name'],
                virtual_data_center_name=cluster['vdc_name'],
                name=cluster['name']
            ),
            api_version=rde_constants.PAYLOAD_VERSION_2_0
        )
        return cluster_entity

    @classmethod
    def get_sample_native_cluster_specification(cls, k8_runtime: str = shared_constants.ClusterEntityKind.NATIVE.value):  # noqa: E501
        """Create apply command cluster specification description.

        :returns: ClusterSpec field description
        :rtype: str
        """
        cluster_spec_field_descriptions = """# Short description of various properties used in this sample cluster configuration
# apiVersion: Represents the payload version of the cluster specification. By default, \"cse.vmware.com/v2.0\" is used.
# kind: The kind of the Kubernetes cluster.
#
# metadata: This is a required section
# metadata.name: Name of the cluster to be created or resized.
# metadata.orgName: The name of the Organization in which cluster needs to be created or managed.
# metadata.virtualDataCenterName: The name of the Organization Virtual data center in which the cluster need to be created or managed.
# metadata.site: VCD site domain name where the cluster should be deployed.
#
# spec: User specification of the desired state of the cluster.
# spec.topology.controlPlane: An optional sub-section for desired control-plane state of the cluster. The properties \"sizingClass\" and \"storageProfile\" can be specified only during the cluster creation phase. These properties will no longer be modifiable in further update operations like \"resize\" and \"upgrade\".
# spec.topology.controlPlane.count: Number of control plane node(s). Only single control plane node is supported.
# spec.topology.controlPlane.sizingClass: The compute sizing policy with which control-plane node needs to be provisioned in a given \"ovdc\". The specified sizing policy is expected to be pre-published to the given ovdc.
# spec.topology.controlPlane.storageProfile: The storage-profile with which control-plane needs to be provisioned in a given \"ovdc\". The specified storage-profile is expected to be available on the given ovdc.
#
# spec.distribution: This is a required sub-section.
# spec.distribution.templateName: Template name based on guest OS, Kubernetes version, and the Weave software version
# spec.distribution.templateRevision: revision number
#
# spec.topology.nfs: Optional sub-section for desired nfs state of the cluster. The properties \"sizingClass\" and \"storageProfile\" can be specified only during the cluster creation phase. These properties will no longer be modifiable in further update operations like \"resize\" and \"upgrade\".
# spec.topology.nfs.count: Nfs nodes can only be scaled-up; they cannot be scaled-down. Default value is 0.
# spec.topology.nfs.sizingClass: The compute sizing policy with which nfs node needs to be provisioned in a given \"ovdc\". The specified sizing policy is expected to be pre-published to the given ovdc.
# spec.topology.nfs.storageProfile: The storage-profile with which nfs needs to be provisioned in a given \"ovdc\". The specified storage-profile is expected to be available on the given ovdc.
#
# spec.settings: This is a required sub-section
# spec.settings.ovdcNetwork: This value is mandatory. Name of the Organization's virtual data center network
# spec.settings.rollbackOnFailure: Optional value that is true by default. On any cluster operation failure, if the value is set to true, affected node VMs will be automatically deleted.
# spec.settings.sshKey: Optional ssh key that users can use to log into the node VMs without explicitly providing passwords.
# spec.settings.network.expose: Optional value that is false by default. Set to true to enable access to the cluster from the external world.
#
# spec.topology.workers: Optional sub-section for the desired worker state of the cluster. The properties \"sizingClass\" and \"storageProfile\" can be specified only during the cluster creation phase. These properties will no longer be modifiable in further update operations like \"resize\" and \"upgrade\". Non uniform worker nodes in the clusters is not yet supported.
# spec.topology.workers.count: number of worker nodes (default value:1) Worker nodes can be scaled up and down.
# spec.topology.workers.sizingClass: The compute sizing policy with which worker nodes need to be provisioned in a given \"ovdc\". The specified sizing policy is expected to be pre-published to the given ovdc.
# spec.topology.workers.storageProfile: The storage-profile with which worker nodes need to be provisioned in a given \"ovdc\". The specified storage-profile is expected to be available on the given ovdc.
#
# status: Current state of the cluster in the server. This is not a required section for any of the operations.\n
"""  # noqa: E501
        metadata = Metadata('cluster_name', 'organization_name',
                            'org_virtual_data_center_name', 'VCD_site')
        status = Status()
        settings = Settings(ovdc_network='ovdc_network_name', ssh_key=None)
        k8_distribution = Distribution(
            template_name='ubuntu-16.04_k8-1.17_weave-2.6.0',
            template_revision=2
        )
        control_plane = ControlPlane(
            count=1,
            sizing_class='Large_sizing_policy_name',
            storage_profile='Gold_storage_profile_name'
        )
        workers = Workers(
            count=2,
            sizing_class='Medium_sizing_policy_name',
            storage_profile='Silver_storage_profile'
        )

        nfs = Nfs(
            count=0,
            sizing_class='Large_sizing_policy_name',
            storage_profile='Platinum_storage_profile_name'
        )

        topology = Topology(
            control_plane=control_plane,
            workers=workers,
            nfs=nfs
        )

        cluster_spec = ClusterSpec(
            topology=topology,
            distribution=k8_distribution,
            settings=settings,
        )

        native_entity_dict = NativeEntity(
            api_version=rde_constants.PAYLOAD_VERSION_2_0,
            metadata=metadata,
            spec=cluster_spec,
            status=status,
            kind=k8_runtime).to_dict()

        # remove status part of the entity dict
        del native_entity_dict['status']

        # Hiding certain portion of the network spec section for
        # Andromeda (CSE 3.1) spec.settings.network is targeted
        # for CSE 3.1.1 to accommodate Antrea as CNI
        # Below line can be deleted post Andromeda (CSE 3.1)
        del native_entity_dict['spec']['settings']['network']['cni']
        del native_entity_dict['spec']['settings']['network']['pods']
        del native_entity_dict['spec']['settings']['network']['services']
        # Hiding the cpu and memory properties from controlPlane and workers
        # for Andromeda (CSE 3.1). Below lines can be deleted once cpu and
        # memory support is added in CSE 3.1.1
        del native_entity_dict['spec']['topology']['controlPlane']['cpu']
        del native_entity_dict['spec']['topology']['controlPlane']['memory']
        del native_entity_dict['spec']['topology']['workers']['cpu']
        del native_entity_dict['spec']['topology']['workers']['memory']

        sample_apply_spec = yaml.dump(native_entity_dict)
        return cluster_spec_field_descriptions + sample_apply_spec



# container-service-extension
# Copyright (c) 2019 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""Utilities to manage CSE metadata on Ovdcs."""

from collections import namedtuple

import pyvcloud.vcd.client as vcd_client
import pyvcloud.vcd.utils as pyvcd_utils
from pyvcloud.vcd.vdc import VDC
import requests

from container_service_extension.common.constants.server_constants import K8S_PROVIDER_KEY  # noqa: E501
from container_service_extension.common.constants.server_constants import K8sProvider  # noqa: E501
from container_service_extension.common.constants.server_constants import PKS_CLUSTER_DOMAIN_KEY  # noqa: E501
from container_service_extension.common.constants.server_constants import PKS_COMPUTE_PROFILE_KEY  # noqa: E501
from container_service_extension.common.constants.server_constants import PKS_PLANS_KEY  # noqa: E501
from container_service_extension.common.constants.shared_constants import RequestKey   # noqa: E501
import container_service_extension.common.utils.pyvcloud_utils as vcd_utils
import container_service_extension.common.utils.server_utils as server_utils
import container_service_extension.exception.exceptions as e
from container_service_extension.logging.logger import SERVER_LOGGER as LOGGER
import container_service_extension.security.context.operation_context as ctx
from container_service_extension.server.pks.pks_cache import PksCache
from container_service_extension.server.pks.pksbroker import PksBroker


def get_ovdc_k8s_provider_metadata(sysadmin_client: vcd_client.Client,
                                   org_name=None, ovdc_name=None, ovdc_id=None,
                                   include_credentials=False,
                                   include_nsxt_info=False):
    """Get k8s provider metadata for an org VDC.

    :param sysadmin_client:
    :param org_name:
    :param ovdc_name:
    :param ovdc_id:
    :param include_credentials:
    :param include_nsxt_info:
    :return:  Dictionary with k8s provider metadata
    """
    vcd_utils.raise_error_if_user_not_from_system_org(sysadmin_client)

    ovdc = vcd_utils.get_vdc(client=sysadmin_client, vdc_name=ovdc_name,
                             vdc_id=ovdc_id, org_name=org_name,
                             is_admin_operation=True)

    all_metadata = pyvcd_utils.metadata_to_dict(ovdc.get_all_metadata())
    k8s_provider = all_metadata.get(K8S_PROVIDER_KEY, K8sProvider.NONE)

    result = {
        K8S_PROVIDER_KEY: k8s_provider
    }

    if k8s_provider == K8sProvider.PKS:
        result.update({k: all_metadata[k] for k in PksCache.get_pks_keys()})  # noqa: E501
        result[PKS_PLANS_KEY] = result[PKS_PLANS_KEY].split(',')

        # Get the credentials from PksCache
        if include_credentials or include_nsxt_info:
            pks_cache = server_utils.get_pks_cache()
            pvdc_info = pks_cache.get_pvdc_info(
                vcd_utils.get_pvdc_id(ovdc))
            if include_credentials:
                # noqa: E501 TODO in case only ovdc_id is provided, we need a way to get org_name
                pks_info = pks_cache.get_pks_account_info(org_name,
                                                          pvdc_info.vc)
                result.update(pks_info.credentials._asdict())
            if include_nsxt_info:
                nsxt_info = pks_cache.get_nsxt_info(pvdc_info.vc)
                result['nsxt'] = nsxt_info

    return result


def get_all_ovdc_with_metadata():
    client = None
    try:
        client = vcd_utils.get_sys_admin_client(api_version=None)
        q = client.get_typed_query(
            vcd_client.ResourceType.ADMIN_ORG_VDC.value,
            query_result_format=vcd_client.QueryResultFormat.RECORDS,
            fields='metadata@SYSTEM:k8s_provider')
        ovdc_records = q.execute()
        return ovdc_records
    finally:
        if client is not None:
            client.logout()


def update_ovdc_k8s_provider_metadata(sysadmin_client: vcd_client.Client,
                                      ovdc_id,
                                      k8s_provider_data=None,
                                      k8s_provider=None):
    """Set the k8s provider metadata for given ovdc.

    :param pyvcloud.vcd.client.Client sysadmin_client:
    :param str ovdc_id:
    :param dict k8s_provider_data:  k8s provider context details
    :param K8sProvider k8s_provider:
    :return:
    """
    vcd_utils.raise_error_if_user_not_from_system_org(sysadmin_client)

    ovdc = vcd_utils.get_vdc(sysadmin_client, vdc_id=ovdc_id)
    ovdc_name = ovdc.get_resource().get('name')
    metadata = {
        K8S_PROVIDER_KEY: k8s_provider or K8sProvider.NONE
    }

    if k8s_provider != K8sProvider.PKS:
        LOGGER.debug(f"Remove existing metadata for ovdc:{ovdc_name}")
        _remove_metadata_from_ovdc(ovdc, PksCache.get_pks_keys())
        LOGGER.debug(f"Updated metadata for {k8s_provider}:"
                     f"{metadata}")
    else:
        k8s_provider_data.pop('username')
        k8s_provider_data.pop('secret')
        k8s_provider_data.pop('nsxt')
        metadata.update(k8s_provider_data)

    # set ovdc metadata into Vcd
    LOGGER.debug(f"On ovdc:{ovdc_name}, setting metadata:{metadata}")
    return ovdc.set_multiple_metadata(metadata,
                                      vcd_client.MetadataDomain.SYSTEM,
                                      vcd_client.MetadataVisibility.PRIVATE)


def construct_pks_context(pks_account_info, pvdc_info=None, nsxt_info=None,
                          pks_compute_profile_name=None, pks_plans='',
                          pks_cluster_domain='',
                          credentials_required=False):
    """Construct PKS context dictionary.

    :param container_service_extension.server.pks.pks_cache.PksAccountInfo pks_account_info: pks connection details  # noqa: E501
    :param container_service_extension.server.pks.pks_cache.PvdcInfo pvdc_info:
        pvdc details including datacenter and cluster.
    :param dict nsxt_info:
    :param str pks_compute_profile_name: name of the compute profile
    :param str pks_plans: comma separated values for pks plans
    :param str pks_cluster_domain:
    :param bool credentials_required: determines if credentials have to
        be part of the resultant dictionary

    :return: pks context details

    :rtype: dict
    """
    pks_ctx = pks_account_info._asdict()
    credentials = pks_ctx.pop('credentials')
    if credentials_required:
        pks_ctx.update(credentials._asdict())
    if pvdc_info:
        pks_ctx.update(pvdc_info._asdict())
    pks_ctx[PKS_COMPUTE_PROFILE_KEY] = '' if not pks_compute_profile_name \
        else pks_compute_profile_name
    pks_ctx[PKS_PLANS_KEY] = '' if not pks_plans else pks_plans
    pks_ctx[PKS_CLUSTER_DOMAIN_KEY] = '' if not pks_cluster_domain \
        else pks_cluster_domain
    pks_ctx['nsxt'] = nsxt_info
    return pks_ctx


def construct_k8s_metadata_from_pks_cache(sysadmin_client: vcd_client.Client,
                                          ovdc_id, org_name, pks_plans,
                                          pks_cluster_domain,
                                          k8s_provider):
    vcd_utils.raise_error_if_user_not_from_system_org(sysadmin_client)

    ctr_prov_context = {
        K8S_PROVIDER_KEY: k8s_provider,
    }
    if k8s_provider == K8sProvider.PKS:
        if not server_utils.is_pks_enabled():
            raise e.CseServerError('CSE is not configured to work with PKS.')

        ovdc = vcd_utils.get_vdc(client=sysadmin_client, vdc_id=ovdc_id,
                                 is_admin_operation=True)
        pks_cache = server_utils.get_pks_cache()
        pvdc_id = vcd_utils.get_pvdc_id(ovdc)
        pvdc_info = pks_cache.get_pvdc_info(pvdc_id)
        if not pvdc_info:
            LOGGER.debug(f"pvdc '{pvdc_id}' is not backed "
                         f"by PKS-managed-vSphere resources")
            raise e.CseServerError(f"VDC '{ovdc.get_resource().get('name')}'"
                                   " is not eligible to provide resources"
                                   " for PKS clusters.")
        pks_account_info = pks_cache.get_pks_account_info(
            org_name, pvdc_info.vc)
        nsxt_info = pks_cache.get_nsxt_info(pvdc_info.vc)

        pks_compute_profile_name = _construct_pks_compute_profile_name(
            sysadmin_client, ovdc_id)
        ctr_prov_context = construct_pks_context(
            pks_account_info=pks_account_info,
            pvdc_info=pvdc_info,
            nsxt_info=nsxt_info,
            pks_compute_profile_name=pks_compute_profile_name,
            pks_plans=pks_plans,
            pks_cluster_domain=pks_cluster_domain,
            credentials_required=True)
    return ctr_prov_context


def _construct_pks_compute_profile_name(sysadmin_client: vcd_client.Client,
                                        vdc_id):
    """Construct pks compute profile name.

    :param pyvcloud.vcd.client.Client sysadmin_client:
    :param str vdc_id: UUID of the vdc in vcd

    :return: pks compute profile name

    :rtype: str
    """
    vcd_utils.raise_error_if_user_not_from_system_org(sysadmin_client)

    vdc = vcd_utils.get_vdc(client=sysadmin_client, vdc_id=vdc_id)
    return f"cp--{vdc_id}--{vdc.name}"


def create_pks_compute_profile(request_data,
                               op_ctx: ctx.OperationContext,
                               pks_context):
    ovdc_id = request_data.get(RequestKey.OVDC_ID)
    org_name = request_data.get(RequestKey.ORG_NAME)
    ovdc_name = request_data.get(RequestKey.OVDC_NAME)
    # Compute profile creation
    pks_compute_profile_name = _construct_pks_compute_profile_name(
        op_ctx.sysadmin_client, ovdc_id)
    pks_compute_profile_description = f"{org_name}--{ovdc_name}--{ovdc_id}"
    pks_az_name = f"az-{ovdc_name}"
    ovdc_rp_name = f"{ovdc_name} ({ovdc_id})"

    compute_profile_params = PksComputeProfileParams(
        pks_compute_profile_name, pks_az_name,
        pks_compute_profile_description,
        pks_context.get('cpi'),
        pks_context.get('datacenter'),
        pks_context.get('cluster'),
        ovdc_rp_name).to_dict()

    LOGGER.debug(f"Creating PKS Compute Profile with name:"
                 f"{pks_compute_profile_name}")

    pksbroker = PksBroker(pks_context, op_ctx)
    try:
        pksbroker.create_compute_profile(**compute_profile_params)
    except e.PksServerError as err:
        if err.status == requests.codes.conflict:
            LOGGER.debug(f"Compute profile name {pks_compute_profile_name}"
                         f" already exists\n{str(err)}")
        else:
            raise


def _remove_metadata_from_ovdc(ovdc: VDC, keys=None):
    if keys is None:
        keys = []
    metadata = pyvcd_utils.metadata_to_dict(ovdc.get_all_metadata())
    for k in keys:
        if k in metadata:
            ovdc.remove_metadata(k, domain=vcd_client.MetadataDomain.SYSTEM)


class PksComputeProfileParams(namedtuple("PksComputeProfileParams",
                                         'cp_name, az_name, description,'
                                         'cpi,datacenter_name, '
                                         'cluster_name, ovdc_rp_name')):
    """Construct PKS ComputeProfile Parameters ."""

    def __str__(self):
        return f"class:{PksComputeProfileParams.__name__}," \
            f" cp_name:{self.cp_name}, az_name:{self.az_name}, " \
            f" description:{self.description}, cpi:{self.cpi}, " \
            f" datacenter_name:{self.datacenter_name}, " \
            f" cluster_name:{self.cluster_name}, " \
            f" ovdc_rp_name:{self.ovdc_rp_name}"

    def to_dict(self):
        return dict(self._asdict())



# container-service-extension
# Copyright (c) 2017 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""Basic utility methods to perform data transformation and file operations."""

import hashlib
import os
import pathlib
import platform
import stat
import sys
from typing import List
import urllib

import click
import pkg_resources
import requests

from container_service_extension.logging.logger import NULL_LOGGER


# chunk size in bytes for file reading
BUF_SIZE = 65536
# chunk size for downloading files
SIZE_1MB = 1024 * 1024

_type_to_string = {
    str: 'string',
    int: 'number',
    bool: 'true/false',
    dict: 'mapping',
    list: 'sequence',
}


class NullPrinter:
    """Callback object which does nothing."""

    def general_no_color(self, msg):
        pass

    def general(self, msg):
        pass

    def info(self, msg):
        pass

    def error(self, msg):
        pass


class ConsoleMessagePrinter(NullPrinter):
    """Callback object to print color coded message on console."""

    def general_no_color(self, msg):
        click.secho(msg)

    def general(self, msg):
        click.secho(msg, fg='green')

    def info(self, msg):
        click.secho(msg, fg='yellow')

    def error(self, msg):
        click.secho(msg, fg='red')


def get_cse_info():
    return {
        'product': 'CSE',
        'description': 'Container Service Extension for VMware vCloud Director',  # noqa: E501
        'version': pkg_resources.require('container-service-extension')[0].version,  # noqa: E501
        'python': platform.python_version()
    }


def prompt_text(text, color='black', hide_input=False):
    click_text = click.style(str(text), fg=color)
    return click.prompt(click_text, hide_input=hide_input, type=str)


def is_environment_variable_enabled(env_var_name):
    """Check if the environment variable is set.

    :param str env_var_name: Name of the environment variable
    :rtype: bool
    """
    return str_to_bool(os.getenv(env_var_name))


def get_duplicate_items_in_list(items):
    """Find duplicate entries in a list.

    :param list items: list of items with possible duplicates.

    :return: the items that occur more than once in input list. Each duplicated
        item will be mentioned only once in the returned list.

    :rtype: list
    """
    seen = set()
    duplicates = set()
    if items:
        for item in items:
            if item in seen:
                duplicates.add(item)
            else:
                seen.add(item)
    return list(duplicates)


def check_keys_and_value_types(dikt, ref_dict, location='dictionary',
                               excluded_keys=None,
                               msg_update_callback=NullPrinter()):
    """Compare a dictionary with a reference dictionary.

    The method ensures that  all keys and value types are the same in the
    dictionaries.

    :param dict dikt: the dictionary to check for validity
    :param dict ref_dict: the dictionary to check against
    :param str location: where this check is taking place, so error messages
        can be more descriptive.
    :param list excluded_keys: list of str, representing the list of key which
        if missing won't raise an exception.
    :param utils.ConsoleMessagePrinter msg_update_callback: Callback object.
    :raises KeyError: if @dikt has missing or invalid keys
    :raises TypeError: if the value of a property in @dikt does not match with
        the value of the same property in @ref_dict
    """
    if excluded_keys is None:
        excluded_keys = []
    ref_keys = set(ref_dict.keys())
    keys = set(dikt.keys())

    missing_keys = ref_keys - keys - set(excluded_keys)

    if missing_keys:
        msg_update_callback.error(
            f"Missing keys in {location}: {missing_keys}")
    bad_value = False
    for k in ref_keys:
        if k not in keys:
            continue
        value_type = type(ref_dict[k])
        if not isinstance(dikt[k], value_type):
            msg_update_callback.error(
                f"{location} key '{k}': value type should be "
                f"'{_type_to_string[value_type]}'")
            bad_value = True

    if missing_keys:
        raise KeyError(f"Missing and/or invalid key in {location}")
    if bad_value:
        raise TypeError(f"Incorrect type for property value(s) in {location}")


def check_python_version(msg_update_callback=NullPrinter()):
    """Ensure that user's Python version >= 3.7.3.

    If the check fails, will exit the python interpreter with error status.

    :param utils.ConsoleMessagePrinter msg_update_callback: Callback object.
    """
    try:
        msg_update_callback.general_no_color(
            "Required Python version: >= 3.7.3\n"
            f"Installed Python version: {sys.version}")
        if sys.version_info < (3, 7, 3):
            raise Exception("Python version should be 3.7.3 or greater")
    except Exception as err:
        msg_update_callback.error(str(err))
        sys.exit(1)


def str_to_bool(s):
    """Convert string boolean values to bool.

    The conversion is case insensitive.

    :param s: input string

    :return: True if val is 'true' otherwise False
    """
    return str(s).lower() == 'true'


def get_sha256(filepath):
    """Get sha256 hash of file as a string.

    :param str filepath: path to file.

    :return: sha256 string for the file.

    :rtype: str
    """
    sha256 = hashlib.sha256()
    with open(filepath, 'rb') as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha256.update(data)
    return sha256.hexdigest()


def check_file_permissions(filename, msg_update_callback=NullPrinter()):
    """Ensure that the file has correct permissions.

    Unix based system:
        Owner - r/w permission
        Other - No access
    Windows:
        No check

    :param str filename: path to file.
    :param utils.ConsoleMessagePrinter msg_update_callback: Callback object.

    :raises Exception: if file has 'x' permissions for Owner or 'rwx'
        permissions for 'Others' or 'Group'.
    """
    if os.name == 'nt':
        return
    err_msgs = []
    file_mode = os.stat(filename).st_mode
    if file_mode & stat.S_IXUSR:
        msg = f"Remove execute permission of the Owner for the file {filename}"
        msg_update_callback.error(msg)
        err_msgs.append(msg)
    if file_mode & stat.S_IROTH or file_mode & stat.S_IWOTH \
            or file_mode & stat.S_IXOTH:
        msg = f"Remove read, write and execute permissions of Others for " \
              f"the file {filename}"
        msg_update_callback.error(msg)
        err_msgs.append(msg)
    if file_mode & stat.S_IRGRP or file_mode & stat.S_IWGRP \
            or file_mode & stat.S_IXGRP:
        msg = f"Remove read, write and execute permissions of Group for the " \
              f"file {filename}"
        msg_update_callback.error(msg)
        err_msgs.append(msg)

    if err_msgs:
        raise IOError(err_msgs)


def download_file(url, filepath, sha256=None, force_overwrite=False,
                  logger=NULL_LOGGER, msg_update_callback=NullPrinter()):
    """Download a file from a url to local filepath.

    Will not overwrite files unless @sha256 is given.
    Recursively creates specified directories in @filepath.

    :param str url: source url.
    :param str filepath: destination filepath.
    :param str sha256: without this argument, if a file already exists at
        @filepath, download will be skipped. If @sha256 matches the file's
        sha256, download will be skipped.
    :param bool force_overwrite: if True, will download the file even if it
        already exists or its SHA hasn't changed.
    :param logging.Logger logger: logger to log with.
    :param utils.ConsoleMessagePrinter msg_update_callback: Callback object.

    :raises HTTPError: if the response has an error status code
    """
    path = pathlib.Path(filepath)
    if not force_overwrite and path.is_file() and \
            (sha256 is None or get_sha256(filepath) == sha256):
        msg = f"Skipping download to '{filepath}' (file already exists)"
        logger.info(msg)
        msg_update_callback.general(msg)
        return

    path.parent.mkdir(parents=True, exist_ok=True)
    msg = f"Downloading file from '{url}' to '{filepath}'..."
    logger.info(msg)
    msg_update_callback.info(msg)
    response = requests.get(url, stream=True,
                            headers={'Cache-Control': 'no-cache'})
    response.raise_for_status()
    with path.open(mode='wb') as f:
        for chunk in response.iter_content(chunk_size=SIZE_1MB):
            f.write(chunk)
    msg = "Download complete"
    logger.info(msg)
    msg_update_callback.general(msg)


def read_data_file(filepath, logger=NULL_LOGGER,
                   msg_update_callback=NullPrinter()):
    """Retrieve file content from local disk as a string.

    :param str filepath: absolute filepath of the file, whose content we want
        to read.
    :param logging.Logger logger: logger to log with.
    :param utils.ConsoleMessagePrinter msg_update_callback: Callback object.

    :return: the contents of the file.

    :rtype: str

    :raises FileNotFoundError: if requested data file cannot be
        found.
    """
    path = pathlib.Path(filepath)
    try:
        contents = path.read_text()
    except FileNotFoundError as err:
        msg_update_callback.error(f"{err}")
        logger.error(f"{err}", exc_info=True)
        raise

    msg = f"Found data file: {path}"
    msg_update_callback.general(msg)
    logger.debug(msg)

    return contents


def flatten_dictionary(input_dict, parent_key='', separator='.'):
    """Flatten a given dictionary with nested dictionaries if any.

    Example: { 'a' : {'b':'c', 'd': {'e' : 'f'}}, 'g' : 'h'} will be flattened
    to {'a.b': 'c', 'a.d.e': 'f', 'g': 'h'}

    This will flatten only the values of dict type.

    :param dict input_dict:
    :param str parent_key: parent key that gets prefixed while forming flattened key  # noqa: E501
    :param str separator: use the separator to form flattened key
    :return: flattened dictionary
    :rtype: dict
    """
    flattened_dict = {}
    for k in input_dict.keys():
        val = input_dict.get(k)
        key_prefix = f"{parent_key}{k}"
        if isinstance(val, dict):
            flattened_dict.update(flatten_dictionary(val, f"{key_prefix}{separator}"))  # noqa: E501
        else:
            flattened_dict.update({key_prefix: val})
    return flattened_dict


def escape_query_filter_expression_value(value):
    value_str = str(value)
    value_str = value_str.replace('(', "\\(")
    value_str = value_str.replace(')', "\\)")
    value_str = value_str.replace(';', "\\;")
    value_str = value_str.replace(',', "\\,")
    return value_str


def construct_filter_string(filters: dict):
    """Construct &-ed filter string from the dict.

    :param dict filters: dictionary containing key and values for the filters
    """
    filter_string = ""
    if filters:
        filter_expressions = []
        for (key, value) in filters.items():
            if key and value:
                filter_exp = f"{key}=={urllib.parse.quote(escape_query_filter_expression_value(value))}"  # noqa: E501
                filter_expressions.append(filter_exp)
        filter_string = ";".join(filter_expressions)
    return filter_string


def extract_id_from_href(href):
    """Extract id from an href.

    'https://vmware.com/api/admin/user/123456' will return 123456

    :param str href: an href

    :return: id
    """
    if not href:
        return None
    if '/' in href:
        return href.split('/')[-1]
    return href


# ToDo: Device a better way to find the max api version
# without converting the strings to float.
# e.g. 5.20 will be smaller than 5.8 if compared as float, which is wrong
def get_max_api_version(api_versions: List[str]) -> str:
    return str(max(float(x) for x in api_versions))



# container-service-extension
# Copyright (c) 2019 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

from container_service_extension.common.constants.shared_constants import CSE_PAGINATION_DEFAULT_PAGE_SIZE  # noqa: E501
from container_service_extension.common.constants.shared_constants import CSE_PAGINATION_FIRST_PAGE_NUMBER  # noqa: E501
from container_service_extension.common.constants.shared_constants import PaginationKey  # noqa: E501
from container_service_extension.common.constants.shared_constants import RequestKey  # noqa: E501
import container_service_extension.common.utils.server_utils as server_utils
from container_service_extension.lib.telemetry.constants import CseOperation
from container_service_extension.lib.telemetry.telemetry_handler import record_user_action_telemetry  # noqa: E501
import container_service_extension.rde.models.common_models as common_models
import container_service_extension.rde.ovdc_service as ovdc_service
import container_service_extension.security.context.operation_context as ctx
import container_service_extension.server.request_handlers.request_utils as request_utils  # noqa: E501


@request_utils.cluster_api_exception_handler
def ovdc_update(data, operation_context: ctx.OperationContext):
    """Request handler for ovdc enable, disable operations.

    Add or remove the respective cluster placement policies to enable or
    disable cluster deployment of a certain kind in the OVDC.

    Required data: k8s_runtime

    :return: Dictionary with org VDC update task href.
    """
    ovdc_spec = common_models.Ovdc(**data[RequestKey.INPUT_SPEC])
    return ovdc_service.update_ovdc(operation_context,
                                    ovdc_id=data[RequestKey.OVDC_ID],
                                    ovdc_spec=ovdc_spec)


@record_user_action_telemetry(cse_operation=CseOperation.OVDC_INFO)
@request_utils.cluster_api_exception_handler
def ovdc_info(data, operation_context: ctx.OperationContext):
    """Request handler for ovdc info operation.

    Required data: ovdc_id

    :return: Dictionary with org VDC k8s provider metadata.
    """
    ovdc_id = data[RequestKey.OVDC_ID]
    return ovdc_service.get_ovdc(operation_context, ovdc_id)


@record_user_action_telemetry(cse_operation=CseOperation.OVDC_LIST)
@request_utils.cluster_api_exception_handler
def ovdc_list(data, operation_context: ctx.OperationContext):
    """Request handler for ovdc list operation.

    :return: List of dictionaries with org VDC k8s runtimes.
    """
    return ovdc_service.list_ovdc(operation_context)


# TODO: Record telemetry in a different telemetry handler
@record_user_action_telemetry(cse_operation=CseOperation.OVDC_LIST)
@request_utils.cluster_api_exception_handler
def org_vdc_list(data, operation_context: ctx.OperationContext):
    """Request handler for org vdc list operation.

    This handler returns a paginated response.
    :return: Dictionary containing paginated response with Org VDC runtime info
    :rtype: dict
    """
    query_params = data.get(RequestKey.QUERY_PARAMS, {})
    page_number = int(query_params.get(PaginationKey.PAGE_NUMBER,
                                       CSE_PAGINATION_FIRST_PAGE_NUMBER))
    page_size = int(query_params.get(PaginationKey.PAGE_SIZE,
                                     CSE_PAGINATION_DEFAULT_PAGE_SIZE))
    result = ovdc_service.list_org_vdcs(operation_context,
                                        page_number=page_number,
                                        page_size=page_size)
    base_uri = data['url']
    return server_utils.create_links_and_construct_paginated_result(
        base_uri,
        result[PaginationKey.VALUES],
        result[PaginationKey.RESULT_TOTAL],
        page_number=page_number,
        page_size=page_size)



# container-service-extension
# Copyright (c) 2021 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause
from dataclasses import dataclass
from enum import Enum
from enum import unique
from typing import Optional

from container_service_extension.common.constants.server_constants import AclAccessLevelId, ExtensionType, MQTT_EXTENSION_URN   # noqa: E501
from container_service_extension.rde.constants import Nss, Vendor

BEHAVIOR_INTERFACE_ID_PREFIX = 'urn:vcloud:behavior-interface'
KUBE_CONFIG_BEHAVIOR_INTERFACE_NAME = 'createKubeConfig'
KUBE_CONFIG_BEHAVIOR_INTERFACE_ID = f"{BEHAVIOR_INTERFACE_ID_PREFIX}:" \
                                    f"{KUBE_CONFIG_BEHAVIOR_INTERFACE_NAME}:" \
                                    f"{Vendor.VMWARE.value}:" \
                                    f"{Nss.KUBERNETES.value}:1.0.0"
KUBE_CONFIG_BEHAVIOR_TYPE_NAME = 'createKubeConfig'
BEHAVIOR_TYPE_ID_PREFIX = 'urn:vcloud:behavior-type'
KUBE_CONFIG_BEHAVIOR_TYPE_ID = f'{BEHAVIOR_TYPE_ID_PREFIX}:' \
                               f'{KUBE_CONFIG_BEHAVIOR_TYPE_NAME}:' \
                               f'{Vendor.CSE.value}:' \
                               f'{Nss.NATIVE_CLUSTER.value}:2.0.0:' \
                               f'{Vendor.VMWARE.value}:' \
                               f'{Nss.KUBERNETES.value}:1.0.0'
CREATE_CLUSTER_BEHAVIOR_NAME = 'createCluster'
CREATE_CLUSTER_BEHAVIOR_INTERFACE_ID = f"{BEHAVIOR_INTERFACE_ID_PREFIX}:" \
                                       f"{CREATE_CLUSTER_BEHAVIOR_NAME}:" \
                                       f"{Vendor.CSE.value}:" \
                                       f"{Nss.KUBERNETES.value}:1.0.0"
UPDATE_CLUSTER_BEHAVIOR_NAME = 'updateCluster'
UPDATE_CLUSTER_BEHAVIOR_INTERFACE_ID = f"{BEHAVIOR_INTERFACE_ID_PREFIX}:" \
                                       f"{UPDATE_CLUSTER_BEHAVIOR_NAME}:" \
                                       f"{Vendor.CSE.value}:" \
                                       f"{Nss.KUBERNETES.value}:1.0.0"
DELETE_CLUSTER_BEHAVIOR_NAME = 'deleteCluster'
DELETE_CLUSTER_BEHAVIOR_INTERFACE_ID = f"{BEHAVIOR_INTERFACE_ID_PREFIX}:" \
                                       f"{DELETE_CLUSTER_BEHAVIOR_NAME}:" \
                                       f"{Vendor.CSE.value}:" \
                                       f"{Nss.KUBERNETES.value}:1.0.0"
DELETE_NFS_NODE_BEHAVIOR_NAME = 'deleteNfsNode'
DELETE_NFS_NODE_BEHAVIOR_ID = f"{BEHAVIOR_INTERFACE_ID_PREFIX}:" \
                              f"{DELETE_NFS_NODE_BEHAVIOR_NAME}:" \
                              f"{Vendor.CSE.value}:" \
                              f"{Nss.KUBERNETES.value}:1.0.0"


@dataclass()
class ExecutionProperties:
    serviceId: str = MQTT_EXTENSION_URN
    # TODO Remove this property once Extensibility team sets the default value.
    invocation_timeout: int = 86400
    actAsToken: bool = True

    def __init__(self, serviceId: str = MQTT_EXTENSION_URN,
                 invocation_timeout: int = 86400, actAsToken: bool = True,
                 **kwargs):
        self.serviceId = serviceId
        self.invocation_timeout = invocation_timeout
        self.actAsToken = actAsToken


@dataclass
class Execution:
    id: str
    type: str = ExtensionType.MQTT.value
    execution_properties: ExecutionProperties = ExecutionProperties()

    def __init__(self, id: str, type: str = ExtensionType.MQTT.value,
                 execution_properties: ExecutionProperties = ExecutionProperties(), **kwargs):  # noqa: E501
        self.id = id
        self.type = type
        self.execution_properties = \
            ExecutionProperties(**execution_properties)\
            if isinstance(execution_properties, dict) else execution_properties


@dataclass
class Behavior:
    name: str
    description: Optional[str] = None
    id: Optional[str] = None
    ref: Optional[str] = None
    execution: Optional[Execution] = None

    def __init__(self, name: str, description: str = None, id: str = None,
                 ref: str = None, execution: Execution = None, **kwargs):
        self.name = name
        self.description = description
        self.id = id
        self.ref = ref
        if execution:
            self.execution = Execution(**execution) if isinstance(execution, dict) else execution  # noqa: E501
        else:
            self.execution = Execution(name)


@dataclass
class BehaviorAclEntry:
    behaviorId: str
    accessLevelId: AclAccessLevelId


@unique
class BehaviorOperation(Enum):
    CREATE_CLUSTER = Behavior(name=CREATE_CLUSTER_BEHAVIOR_NAME,
                              description='Creates native cluster',
                              id=CREATE_CLUSTER_BEHAVIOR_INTERFACE_ID)
    UPDATE_CLUSTER = Behavior(name=UPDATE_CLUSTER_BEHAVIOR_NAME,
                              description='Updates native cluster',
                              id=UPDATE_CLUSTER_BEHAVIOR_INTERFACE_ID)
    DELETE_CLUSTER = Behavior(name=DELETE_CLUSTER_BEHAVIOR_NAME,
                              description='Deletes native cluster',
                              id=DELETE_CLUSTER_BEHAVIOR_INTERFACE_ID)
    GET_KUBE_CONFIG = Behavior(name=KUBE_CONFIG_BEHAVIOR_INTERFACE_NAME,
                               id=KUBE_CONFIG_BEHAVIOR_TYPE_ID,
                               ref=KUBE_CONFIG_BEHAVIOR_INTERFACE_ID)
    DELETE_NFS_NODE = Behavior(name=DELETE_NFS_NODE_BEHAVIOR_NAME,
                               id=DELETE_NFS_NODE_BEHAVIOR_ID,
                               ref=DELETE_NFS_NODE_BEHAVIOR_ID)


@unique
class BehaviorAcl(Enum):
    CREATE_CLUSTER_ACL = BehaviorAclEntry(CREATE_CLUSTER_BEHAVIOR_INTERFACE_ID,
                                          AclAccessLevelId.AccessLevelReadWrite)  # noqa: E501
    UPDATE_CLUSTER_ACL = BehaviorAclEntry(UPDATE_CLUSTER_BEHAVIOR_INTERFACE_ID,
                                          AclAccessLevelId.AccessLevelReadWrite)  # noqa: E501
    DELETE_CLUSTER_ACL = BehaviorAclEntry(DELETE_CLUSTER_BEHAVIOR_INTERFACE_ID,
                                          AclAccessLevelId.AccessLevelFullControl)  # noqa: E501
    KUBE_CONFIG_ACL = BehaviorAclEntry(KUBE_CONFIG_BEHAVIOR_INTERFACE_ID,
                                       AclAccessLevelId.AccessLevelReadOnly)
    DELETE_NFS_NODE_ACL = BehaviorAclEntry(DELETE_NFS_NODE_BEHAVIOR_ID,
                                           AclAccessLevelId.AccessLevelReadWrite)  # noqa: E501


@dataclass()
class BehaviorError:
    majorErrorCode: str = '400'
    minorErrorCode: Optional[str] = None
    message: Optional[str] = None


@unique
class BehaviorTaskStatus(Enum):
    RUNNING = 'running'
    SUCCESS = 'success'
    ERROR = 'error'
    ABORTED = 'aborted'



# container-service-extension
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""Constants used only in the CSE CLI."""

from enum import Enum
from enum import unique

import container_service_extension.common.constants.shared_constants as shared_constants  # noqa: E501


# Client environment variables
ENV_CSE_CLIENT_WIRE_LOGGING = 'CSE_CLIENT_WIRE_LOGGING'
ENV_CSE_TKG_PLUS_ENABLED = 'CSE_TKG_PLUS_ENABLED'


# if cse_server_running key is set to false in profiles.yaml, CSE CLI can
# only be used to work with TKG clusters. This key is set when the first call
# to CSE server is made. If CSE server starts running after the first call
# fails, a re-login is needed to reset the key in profiles.yaml
CSE_SERVER_RUNNING = 'cse_server_running'


TKG_RESPONSE_MESSAGES_BY_STATUS_CODE = {
    403: "User doesn't have required rights to perform the operation",
    500: "Unexpected error occurred"
}

# Fields for cluster acl request
CLUSTER_ACL_UPDATE_REQUEST_FIELDS = \
    [shared_constants.AccessControlKey.ACCESS_LEVEL_ID,
     shared_constants.AccessControlKey.MEMBER_ID,
     shared_constants.AccessControlKey.USERNAME]

# CLI pagination constant to be consistent with UI pagination
CLI_ENTRIES_PER_PAGE = 10


@unique
class CLIOutputKey(str, Enum):
    """Keys for displaying Cluster list output."""

    CLUSTER_NAME = "Name"
    VDC = "VDC"
    ORG = "Org"
    K8S_RUNTIME = "K8s Runtime"
    K8S_VERSION = "K8s Version"
    STATUS = "Status"
    OWNER = "Owner"


@unique
class GroupKey(str, Enum):
    CLUSTER = 'cluster'
    NODE = 'node'
    OVDC = 'ovdc'
    PKS = 'pks'
    SYSTEM = 'system'
    TEMPLATE = 'template'
    VERSION = 'version'


@unique
class CommandNameKey(str, Enum):
    CONFIG = 'config'
    CREATE = 'create'
    DELETE = 'delete'
    UPGRADE = 'upgrade'
    UPGRADE_PLAN = 'upgrade-plan'
    INFO = 'info'
    NODE = 'node'
    ENABLE = 'enable'
    DISABLE = 'disable'


@unique
class TKGEntityFilterKey(str, Enum):
    """Keys to filter TKG cluster entities in CSE CLI.

    Below Keys are commonly used filters. An entity can be filtered by any of
    its properties.
    """

    CLUSTER_NAME = 'entity.metadata.name'
    VDC_NAME = 'entity.metadata.virtualDataCenterName'
    ORG_NAME = 'org.name'


@unique
class TKGRequestHeaderKey(str, Enum):
    """Header keys for tkgCluster requests."""

    AUTHORIZATION = 'Authorization'
    X_VCLOUD_AUTHORIZATION = 'x-vcloud-authorization'
    ACCEPT = 'Accept'
    X_VMWARE_VCLOUD_TENANT_CONTEXT = "x-vmware-vcloud-tenant-context"



# container-service-extension
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

import functools
import json
from typing import List, Tuple, Union

from pyvcloud.vcd.client import ApiVersion
from requests.exceptions import HTTPError

import container_service_extension.common.constants.shared_constants as shared_constants  # noqa: E501
from container_service_extension.common.constants.shared_constants import CSE_PAGINATION_DEFAULT_PAGE_SIZE, PaginationKey  # noqa: E501
from container_service_extension.common.constants.shared_constants import CSE_PAGINATION_FIRST_PAGE_NUMBER  # noqa: E501
from container_service_extension.common.constants.shared_constants import HttpResponseHeader  # noqa: E501
from container_service_extension.common.constants.shared_constants import RequestMethod  # noqa: E501
import container_service_extension.common.utils.core_utils as utils
import container_service_extension.exception.exceptions as cse_exception
from container_service_extension.exception.minor_error_codes import MinorErrorCode  # noqa: E501
from container_service_extension.lib.cloudapi.cloudapi_client import CloudApiClient  # noqa: E501
from container_service_extension.lib.cloudapi.constants import CloudApiResource
from container_service_extension.lib.cloudapi.constants import CloudApiVersion
from container_service_extension.logging.logger import SERVER_LOGGER as LOGGER
import container_service_extension.rde.constants as def_constants
from container_service_extension.rde.models.abstractNativeEntity import AbstractNativeEntity  # noqa: E501
from container_service_extension.rde.models.common_models import DefEntity
from container_service_extension.rde.models.common_models import DefEntityType
from container_service_extension.rde.models.common_models import GenericClusterEntity  # noqa: E501
from container_service_extension.rde.models.common_models import TKGEntity
import container_service_extension.rde.utils as def_utils


def handle_entity_service_exception(func):
    """Decorate to trap exceptions and process them.

    Raise errors of type HTTPError as DefEntityServiceError
    and raise the custom error. Re-raise any other exception as it is.

    This decorator should be applied only on method of entity_service.py

    :param method func: decorated function

    :return: reference to the function that executes the decorated function
        and traps exceptions raised by it.
    """
    @functools.wraps(func)
    def exception_handler_wrapper(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
        except HTTPError as error:
            response_dict = json.loads(error.response.text)
            error_message = response_dict.get('message')
            LOGGER.error(error_message)
            raise cse_exception.DefEntityServiceError(
                error_message=error_message,
                minor_error_code=MinorErrorCode.DEFAULT_ERROR_CODE)
        except Exception as error:
            LOGGER.error(error)
            raise error
        return result
    return exception_handler_wrapper


class DefEntityService:
    """Manages lifecycle of entities.

    TODO Add API version check at the appropriate place. This class needs to
     be used if and only if vCD API version >= v35.
    """

    def __init__(self, cloudapi_client: CloudApiClient):
        def_utils.raise_error_if_def_not_supported(cloudapi_client)
        self._cloudapi_client = cloudapi_client

    @handle_entity_service_exception
    def create_entity(self, entity_type_id: str, entity: DefEntity,
                      tenant_org_context: str = None,
                      delete_status_from_payload=True,
                      is_request_async=False) -> Union[dict, Tuple[dict, dict]]:  # noqa: E501
        """Create defined entity instance of an entity type.

        :param str entity_type_id: ID of the DefEntityType
        :param DefEntity entity: Defined entity instance
        :param str tenant_org_context:
        :param bool delete_status_from_payload: should delete status from payload?  # noqa: E501
        :param bool is_request_async: The request is intended to be asynchronous
            if this flag is set, href of the task is returned in addition to
            the response body

        :return: created entity or created entity with response headers
        :rtype: Union[dict, Tuple[dict, dict]]
        """
        additional_request_headers = {}
        if tenant_org_context:
            additional_request_headers['x-vmware-vcloud-tenant-context'] = tenant_org_context  # noqa: E501

        payload: dict = entity.to_dict()
        if delete_status_from_payload:
            payload.get('entity', {}).pop('status', None)

        resource_url_relative_path = f"{CloudApiResource.ENTITY_TYPES}/{entity_type_id}"  # noqa: E501
        # response will be a tuple (response_body, response_header) if
        # is_request_async is true. Else, it will be just response_body
        response = self._cloudapi_client.do_request(
            method=RequestMethod.POST,
            cloudapi_version=CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=resource_url_relative_path,
            payload=payload,
            additional_request_headers=additional_request_headers,
            return_response_headers=is_request_async)

        if is_request_async:
            # if request is async, return the location header as well
            # TODO: Use the Http response status code to decide which
            #   header name to use for task_href
            #   202 - location header,
            #   200 - xvcloud-task-location needs to be used
            return response[0], response[1][HttpResponseHeader.LOCATION.value]
        return response

    @handle_entity_service_exception
    def list_entities_by_entity_type(self, vendor: str, nss: str, version: str,
                                     filters: dict = None) -> List[DefEntity]:
        """List entities of a given entity type.

        vCD's behavior when invalid filter keys are passed:
            * It will throw a 400 if invalid first-level filter keys are passed
            Valid keys : [name, id, externalId, entityType, entity, state].
            * It will simply return empty list on passing any invalid nested
             properties.

        :param str vendor: Vendor of the entity type
        :param str nss: nss of the entity type
        :param str version: version of the entity type
        :param dict filters: Key-value pairs representing filter options
        :return: List of entities of that entity type
        :rtype: Generator[DefEntity, None, None]
        """
        filter_string = utils.construct_filter_string(filters)
        page_num = 0
        while True:
            page_num += 1
            query_string = f"page={page_num}&sortAsc=name"
            if filter_string:
                query_string = f"filter={filter_string}&{query_string}"
            response_body = self._cloudapi_client.do_request(
                method=RequestMethod.GET,
                cloudapi_version=CloudApiVersion.VERSION_1_0_0,
                resource_url_relative_path=f"{CloudApiResource.ENTITIES}/"
                                           f"{CloudApiResource.ENTITY_TYPES_TOKEN}/"  # noqa: E501
                                           f"{vendor}/{nss}/{version}?{query_string}")  # noqa: E501
            if len(response_body['values']) == 0:
                break
            for entity in response_body['values']:
                yield DefEntity(**entity)

    @handle_entity_service_exception
    def get_entities_per_page_by_entity_type(self, vendor: str, nss: str, version: str,  # noqa: E501
                                             filters: dict = None, page_number: int = CSE_PAGINATION_FIRST_PAGE_NUMBER,  # noqa: E501
                                             page_size: int = CSE_PAGINATION_DEFAULT_PAGE_SIZE):  # noqa: E501
        """List all the entities per page and entity type.

        :param str vendor: entity type vendor name
        :param str nss: entity type namespace
        :param str version: entity type version
        :param dict filters: additional filters
        :param int page_number: page to return
        :param int page_size: number of records per page
        :rtype: Generator[(List[DefEntity], int), None, None]
        """
        filter_string = utils.construct_filter_string(filters)
        query_string = f"page={page_number}&pageSize={page_size}"
        if filter_string:
            query_string = f"filter={filter_string}&{query_string}"
        response_body = self._cloudapi_client.do_request(
            method=RequestMethod.GET,
            cloudapi_version=CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=f"{CloudApiResource.ENTITIES}/"
                                       f"{CloudApiResource.ENTITY_TYPES_TOKEN}/"  # noqa: E501
                                       f"{vendor}/{nss}/{version}?{query_string}")  # noqa: E501
        result = {}
        entity_list: List[DefEntity] = []
        for v in response_body['values']:
            entity_list.append(DefEntity(**v))
        result[PaginationKey.RESULT_TOTAL] = int(response_body['resultTotal'])
        result[PaginationKey.VALUES] = entity_list
        return result

    @handle_entity_service_exception
    def list_entities_by_interface(self, vendor: str, nss: str, version: str):
        """List entities of a given interface.

        An interface is uniquely identified by properties vendor, nss and
        version.

        :param str vendor: Vendor of the interface
        :param str nss: nss of the interface
        :param str version: version of the interface
        :return: Generator of entities of that interface type
        :rtype: Generator[DefEntity, None, None]
        """
        # TODO Yet to be verified. Waiting for the build from Extensibility
        #  team.
        page_num = 0
        while True:
            page_num += 1
            response_body = self._cloudapi_client.do_request(
                method=RequestMethod.GET,
                cloudapi_version=CloudApiVersion.VERSION_1_0_0,
                resource_url_relative_path=f"{CloudApiResource.ENTITIES}/"
                                           f"{CloudApiResource.INTERFACES}/{vendor}/{nss}/{version}?"  # noqa: E501
                                           f"page={page_num}")
            if len(response_body['values']) == 0:
                break
            for entity in response_body['values']:
                yield DefEntity(**entity)

    def get_all_entities_per_page_by_interface(self, vendor: str, nss: str, version: str,  # noqa: E501
                                               filters: dict = None,
                                               page_number: int = CSE_PAGINATION_FIRST_PAGE_NUMBER,  # noqa: E501
                                               page_size: int = CSE_PAGINATION_DEFAULT_PAGE_SIZE):  # noqa: E501
        """Get a page of entities belonging to an interface.

        An interface is uniquely identified by properties vendor, nss and
        version.

        :param str vendor: Vendor of the interface
        :param str nss: nss of the interface
        :param str version: version of the interface
        :param dict filters:
        :param int page_number:
        :param int page_size:

        :return: Generator of entities of that interface type
        :rtype: Generator[List, None, None]
        """
        filter_string = utils.construct_filter_string(filters)
        query_string = f"page={page_number}&pageSize={page_size}"
        if filter_string:
            query_string = f"filter={filter_string}&{query_string}"
        response_body = self._cloudapi_client.do_request(
            method=RequestMethod.GET,
            cloudapi_version=CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=f"{CloudApiResource.ENTITIES}/"
                                       f"{CloudApiResource.INTERFACES}/"
                                       f"{vendor}/{nss}/{version}?{query_string}")  # noqa: E501
        result = {}
        entity_list = []
        for entity in response_body['values']:
            entity_list.append(GenericClusterEntity(**entity))
        result[PaginationKey.RESULT_TOTAL] = int(response_body['resultTotal'])
        result[PaginationKey.PAGE_COUNT] = int(response_body['pageCount'])
        result[PaginationKey.PAGE_NUMBER] = page_number
        result[PaginationKey.PAGE_SIZE] = page_size
        result[PaginationKey.VALUES] = entity_list
        return result

    @handle_entity_service_exception
    def update_entity(self, entity_id: str, entity: DefEntity,
                      invoke_hooks=False,
                      is_request_async=False) -> Union[DefEntity, Tuple[DefEntity, dict]]:  # noqa: E501
        """Update entity instance.

        :param str entity_id: Id of the entity to be updated.
        :param DefEntity entity: Modified entity to be updated.
        :param bool invoke_hooks: Value indicating whether
             hook-based-behaviors need to be triggered or not.
        :param bool is_request_async: The request is intended to be
            asynchronous if this flag is set, href of the task is returned
            in addition to the response body

        :return: Updated entity or Updated entity and response headers
        :rtype: Union[DefEntity, Tuple[DefEntity, dict]]
        """
        resource_url_relative_path = f"{CloudApiResource.ENTITIES}/{entity_id}"
        vcd_api_version = self._cloudapi_client.get_api_version()
        # TODO Float conversions must be changed to Semantic versioning.
        # TODO: Also include any persona having Administrator:FullControl
        #  on CSE:nativeCluster
        if float(vcd_api_version) >= float(ApiVersion.VERSION_36.value) and \
                self._cloudapi_client.is_sys_admin and not invoke_hooks:
            resource_url_relative_path += f"?invokeHooks={str(invoke_hooks).lower()}"  # noqa: E501

        payload: dict = entity.to_dict()

        # Prevent users with rights <= EDIT/VIEW on CSE:NATIVECLUSTER from
        # updating "private" property of RDE "status" section
        # TODO: Replace sys admin check with FULL CONTROL rights check on
        #  CSE:NATIVECLUSTER. Users with no FULL CONTROL rights cannot update
        #  private property of entity->status.
        if not self._cloudapi_client.is_sys_admin:
            payload.get('entity', {}).get('status', {}).pop('private', None)

        if is_request_async:
            # if request is async, return the task href in
            # x_vmware_vcloud_task_location header
            # TODO: Use the Http response status code to decide which
            #   header name to use for task_href
            #   202 - location header,
            #   200 - xvcloud-task-location needs to be used
            response_body, headers = self._cloudapi_client.do_request(
                method=RequestMethod.PUT,
                cloudapi_version=CloudApiVersion.VERSION_1_0_0,
                resource_url_relative_path=resource_url_relative_path,
                payload=payload,
                return_response_headers=is_request_async)
            return DefEntity(**response_body), headers[HttpResponseHeader.X_VMWARE_VCLOUD_TASK_LOCATION.value]  # noqa: E501
        else:
            response_body = self._cloudapi_client.do_request(
                method=RequestMethod.PUT,
                cloudapi_version=CloudApiVersion.VERSION_1_0_0,
                resource_url_relative_path=resource_url_relative_path,
                payload=payload)
            return DefEntity(**response_body)

    @handle_entity_service_exception
    def get_entity(self, entity_id: str) -> DefEntity:
        """Get the defined entity given entity id.

        :param str entity_id: Id of the entity.
        :return: Details of the entity.
        :rtype: DefEntity
        """
        response_body = self._cloudapi_client.do_request(
            method=RequestMethod.GET,
            cloudapi_version=CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=f"{CloudApiResource.ENTITIES}/"
                                       f"{entity_id}")
        return DefEntity(**response_body)

    @handle_entity_service_exception
    def get_tkg_or_def_entity(self, entity_id: str) -> DefEntity:
        """Get the tkg or def entity given entity id.

        :param str entity_id: Id of the entity.

        :return: Details of the entity.
        :rtype: DefEntity
        """
        response_body = self._cloudapi_client.do_request(
            method=RequestMethod.GET,
            cloudapi_version=CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=f"{CloudApiResource.ENTITIES}/"
                                       f"{entity_id}")
        entity: dict = response_body['entity']
        entity_kind: dict = entity['kind']
        if entity_kind in [shared_constants.ClusterEntityKind.NATIVE.value,
                           shared_constants.ClusterEntityKind.TKG_PLUS.value]:
            return DefEntity(**response_body)
        elif entity_kind == shared_constants.ClusterEntityKind.TKG_S.value:
            return TKGEntity(**entity)
        raise Exception("Invalid cluster kind.")

    @handle_entity_service_exception
    def create_acl_for_entity(self,
                              id: str,
                              grant_type: str,
                              access_level_id: str,
                              member_id: str):
        """Create ACL Rule for the Entity.

        :param str id: Id of the Entity
        :param str grant_type: if acl grant is based on memberships or
            entitlements
        :param str access_level_id: level of access which the
            subject will be granted.
        :param str member_id: member id, this access control grant applies to
        """
        acl_details = {
            'grantType': grant_type,
            'accessLevelId': access_level_id,
            'memberId': member_id
        }

        self._cloudapi_client.do_request(
            method=RequestMethod.POST,
            cloudapi_version=CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=f"{CloudApiResource.ENTITIES}/"
                                       f"{id}/"
                                       f"{CloudApiResource.ACL}/",
            payload=acl_details)
        return

    @handle_entity_service_exception
    def get_native_rde_by_name_and_rde_version(self, name: str, version: str,
                                               filters: dict = None) -> DefEntity:  # noqa: E501
        """Get native RDE given its name and RDE version.

        This function is used commonly by CSE CLI client and CSE server

        :param str name: Name of the native cluster.
        :param str version: RDE version
        :param dict filters: dictionary representing filters
        :rtype: DefEntity
        :return: Native cluster RDE
        """
        if not filters:
            filters = {}
        filters[def_constants.RDEFilterKey.NAME.value] = name
        native_entity_type: DefEntityType = \
            def_utils.get_rde_metadata(version)[def_constants.RDEMetadataKey.ENTITY_TYPE]  # noqa: E501
        for entity in \
            self.list_entities_by_entity_type(vendor=native_entity_type.vendor,
                                              nss=native_entity_type.nss,
                                              version=native_entity_type.version,  # noqa: E501
                                              filters=filters):
            return entity

    @handle_entity_service_exception
    def delete_entity(self, entity_id: str, invoke_hooks: bool = False, is_request_async=False) -> Union[dict, Tuple[dict, dict]]:  # noqa: E501
        """Delete the defined entity.

        :param str entity_id: Id of the entity.
        :param bool invoke_hooks: set to true if hooks need to be invoked
        :param bool is_request_async: The request is intended to be
            asynchronous if this flag is set, href of the task is returned
            in addition to the response body

        :return: response body or response body and response headers
        :rtype: Union[dict, Tuple[dict, dict]]
        """
        # response will be a tuple (response_body, response_header) if
        # is_request_async is true. Else, it will be just response_body
        vcd_api_version = self._cloudapi_client.get_api_version()
        resource_url_relative_path = f"{CloudApiResource.ENTITIES}/{entity_id}"

        # TODO: Also include any persona having Administrator:FullControl
        #  on CSE:nativeCluster
        if float(vcd_api_version) >= float(ApiVersion.VERSION_36.value) and \
                self._cloudapi_client.is_sys_admin and not invoke_hooks:
            resource_url_relative_path += f"?invokeHooks={str(invoke_hooks).lower()}"  # noqa: E501

        response = self._cloudapi_client.do_request(
            method=RequestMethod.DELETE,
            cloudapi_version=CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=resource_url_relative_path,  # noqa: E501
            return_response_headers=is_request_async)
        if is_request_async:
            # if request is async, return the location header as well
            # TODO: Use the Http response status code to decide which
            #   header name to use for task_href
            #   202 - location header,
            #   200 - xvcloud-task-location needs to be used
            return response[0], response[1][HttpResponseHeader.LOCATION.value]
        return response

    @handle_entity_service_exception
    def resolve_entity(self, entity_id: str, entity_type_id: str = None) -> DefEntity:  # noqa: E501
        """Resolve the entity.

        Validates the entity against the schema. Based on the result, entity
        state will be either changed to "RESOLVED" (or) "RESOLUTION ERROR".

        :param str entity_id: Id of the entity
        :param str entity_type_id: Entity type ID of the entity being resolved
        :return: Defined entity with its state updated.
        :rtype: DefEntity
        """
        if not entity_type_id:
            rde: DefEntity = self.get_entity(entity_id)
            entity_type_id = rde.entityType
        response_body = self._cloudapi_client.do_request(
            method=RequestMethod.POST,
            cloudapi_version=CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=f"{CloudApiResource.ENTITIES}/"
                                       f"{entity_id}/{CloudApiResource.ENTITY_RESOLVE}")  # noqa: E501
        msg = response_body[def_constants.DEF_ERROR_MESSAGE_KEY]
        del response_body[def_constants.DEF_ERROR_MESSAGE_KEY]
        entity = DefEntity(entityType=entity_type_id, **response_body)
        # TODO: Just record the error message; revisit after HTTP response code
        # is good enough to decide if exception should be thrown or not
        if entity.state != def_constants.DEF_RESOLVED_STATE:
            LOGGER.error(msg)
        return entity

    def is_native_entity(self, entity_id: str):
        """."""
        response_body = self._cloudapi_client.do_request(
            method=RequestMethod.GET,
            cloudapi_version=CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=f"{CloudApiResource.ENTITIES}/"
                                       f"{entity_id}")
        return def_constants.DEF_NATIVE_ENTITY_TYPE_NSS in response_body['entityType']  # noqa: E501

    @handle_entity_service_exception
    def upgrade_entity(self, entity_id: str,
                       upgraded_native_entity: AbstractNativeEntity,
                       target_entity_type_id: str):
        """Upgrade entity type of the entity to the specified one.

        :param str entity_id: ID of the entity to upgrade
        :param str upgraded_native_entity: dataclass representing the native
            entity with upgraded fields.
        :param str target_entity_type_id: target entity type version to which
            the defined entity should be upgraded to.
        :return: DefEntity representing the upgraded defined entity
        :rtype: DefEntity
        """
        rde = self.get_entity(entity_id)
        rde.entity = upgraded_native_entity

        # Update only the entityType property
        rde.entityType = target_entity_type_id

        return self.update_entity(entity_id, rde, invoke_hooks=False)



# container-service-extension
# Copyright (c) 2019 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause
from pyvcloud.vcd import utils
import pyvcloud.vcd.exceptions as vcd_exceptions

import container_service_extension.client.cse_client.api_35.ovdc_api as ovdc_api_v35  # noqa: E501
import container_service_extension.client.utils as client_utils
from container_service_extension.common.utils.pyvcloud_utils import get_vdc
import container_service_extension.rde.models.common_models as common_models


class PolicyBasedOvdc:
    def __init__(self, client):
        self.client = client
        self._ovdc_api = ovdc_api_v35.OvdcApi(self.client)

    def list_ovdc(self):
        for ovdc_list, has_more_results in self._ovdc_api.get_all_ovdcs():
            value_field_to_display_field = {
                'ovdc_name': 'Name',
                'ovdc_id': 'ID',
                'k8s_runtime': 'K8s Runtime'
            }
            yield client_utils.filter_columns(ovdc_list, value_field_to_display_field), has_more_results  # noqa: E501

    def update_ovdc(self, ovdc_name, k8s_runtime, enable=True, org_name=None,
                    remove_cp_from_vms_on_disable=False):
        """Enable/Disable ovdc for k8s for the given k8s provider.

        :param str ovdc_name: Name of org VDC to update
        :param List[str] k8s_runtime: k8s_runtime of the k8s provider to
        enable / disable for the ovdc
        :param bool enable: If set to True will enable the vdc for the
            paricular k8s_runtime else if set to False, K8 support on
            the vdc will be disabled.
        :param str org_name: Name of org that @ovdc_name belongs to
        :param bool remove_cp_from_vms_on_disable: If set to True and
            enable is False, then all the vms in the ovdc having policies for
            the k8s_runtime is deleted.

        :rtype: dict
        """
        ovdc = get_vdc(self.client, vdc_name=ovdc_name, org_name=org_name,
                       is_admin_operation=True)
        ovdc_id = utils.extract_id(ovdc.get_resource().get('id'))
        curr_ovdc = self._ovdc_api.get_ovdc(ovdc_id)
        runtimes = curr_ovdc.k8s_runtime
        for k in k8s_runtime:
            if enable:
                if k in runtimes:
                    raise Exception(f"OVDC {ovdc_name} already enabled for {k}") # noqa: E501
                runtimes.append(k)
            else:
                if k not in runtimes:
                    raise Exception(f"OVDC {ovdc_name} already disabled for {k}") # noqa: E501
                runtimes.remove(k)
        updated_ovdc = common_models.Ovdc(
            k8s_runtime=runtimes,
            org_name=org_name,
            remove_cp_from_vms_on_disable=remove_cp_from_vms_on_disable)
        return self._ovdc_api.update_ovdc(ovdc_id, updated_ovdc)

    def info_ovdc(self, ovdc_name, org_name):
        """Disable ovdc for given kubernetes runtime.

        :param str ovdc_name: Name of the org VDC to be enabled
        :param str org_name: Name of org that @ovdc_name belongs to

        :rtype: dict
        """
        raise NotImplementedError("OVDC info functionality is not supported "
                                  "for the installed CSE version.")

    # TODO(compute-policy for v35): Revisit after decision on api v35 support
    def update_ovdc_compute_policies(self, ovdc_name, org_name,
                                     compute_policy_name, action,
                                     remove_compute_policy_from_vms):
        """Update an ovdc's compute policies.

        :param str ovdc_name: Name of org VDC to update
        :param str org_name: Name of org that @ovdc_name belongs to
        :param str compute_policy_name: Name of compute policy to add or remove
        :param ComputePolicyAction action:

        :rtype: dict
        """
        msg = "Operation not supported; Under implementation"
        raise vcd_exceptions.OperationNotSupportedException(msg)

    # TODO(compute-policy for v35): Revisit after decision on api v35 support
    def list_ovdc_compute_policies(self, ovdc_name, org_name):
        """List an ovdc's compute policies.

        :param str ovdc_name: Name of org VDC to update
        :param str org_name: Name of org that @ovdc_name belongs to

        :rtype: dict
        """
        msg = "Operation not supported; Under implementation"
        raise vcd_exceptions.OperationNotSupportedException(msg)



# container-service-extension
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

import pyvcloud.vcd.client as vcd_client

from container_service_extension.client.cse_client.cse_client import CseClient
import container_service_extension.common.constants.shared_constants as shared_constants  # noqa: E501


class PksOvdcApi(CseClient):
    def __init__(self, client: vcd_client.Client):
        super().__init__(client)
        self._uri = f"{self._uri}/{shared_constants.PKS_URL_FRAGMENT}"
        self._org_vdcs_uri = f"{self._uri}/orgvdcs"
        self._ovdc_uri = f"{self._uri}/ovdc"

    def get_all_ovdcs(self, filters=None):
        if filters is None:
            filters = {}
        filters[shared_constants.PaginationKey.PAGE_SIZE.value] = self._request_page_size  # noqa: E501

        return self.iterate_results(self._org_vdcs_uri, filters=filters)

    def get_ovdc(self, ovdc_id):
        uri = f"{self._ovdc_uri}/{ovdc_id}"
        response = self.do_request(
            uri=uri,
            method=shared_constants.RequestMethod.GET,
            accept_type='application/json')
        return self.process_response(response)

    def update_ovdc_by_ovdc_id(self, ovdc_id, k8s_provider, ovdc_name=None,
                               org_name=None, pks_plan=None, pks_cluster_domain=None):  # noqa: E501
        payload = {
            shared_constants.RequestKey.OVDC_ID: ovdc_id,
            shared_constants.RequestKey.OVDC_NAME: ovdc_name,
            shared_constants.RequestKey.ORG_NAME: org_name,
            shared_constants.RequestKey.K8S_PROVIDER: k8s_provider,
            shared_constants.RequestKey.PKS_PLAN_NAME: pks_plan,
            shared_constants.RequestKey.PKS_CLUSTER_DOMAIN: pks_cluster_domain
        }
        uri = f"{self._ovdc_uri}/{ovdc_id}"
        response = self.do_request(
            uri=uri,
            method=shared_constants.RequestMethod.PUT,
            accept_type='application/json',
            media_type='application/json',
            payload=payload)
        return self.process_response(response)
